/* -*- Mode: C++; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is TclXPCOM.
 * 
 * The Initial Developer of the Original Code is Mark Follett.
 * Portions created by Mark Follett are Copyright (C) 2001-2002
 * Mark Follett.  All Rights Reserved.
 * 
 * Contributor(s):
 *     Mark Follett <mef123@myrealbox.com> (Original Author)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 */

/**
 *  @file TclXPCOMError.cpp
 *  Contains functions for handling XPCOM result codes in Tcl.
 */

#include "TclXPCOMPrivate.h"


/**
 *  Sets an interpreter's 'errorCode' variable to contain an
 *  XPCOM result code. The format consists of a list whose first
 *  element is the string 'XPCOM' and whose second element is the
 *  result code.
 */
void
TclXPCOM_SetErrorCode(
    Tcl_Interp *interp, ///< (in) interpreter in which to set 'errorCode'
    nsresult    res     ///< (in) XPCOM result code
)
{
    Tcl_Obj *errorCode;

    NS_PRECONDITION(interp, "null pointer");

    errorCode = Tcl_NewListObj(0, NULL);
    Tcl_ListObjAppendElement(NULL, errorCode, Tcl_NewStringObj("XPCOM", -1));
    Tcl_ListObjAppendElement(NULL, errorCode, Tcl_NewLongObj((long)res));
    Tcl_SetObjErrorCode(interp, errorCode);
}


/**
 *  Attempts to retrieve an XPCOM result code from an interpreter's
 *  'errorCode' variable.
 *
 *  @retval 1 - if the interp's errorCode contains an XPCOM result code
 *  @retval 0 - if it doesn't
 */
int
TclXPCOM_GetErrorCode(
    Tcl_Interp *interp, ///< (in)  interpreter to search
    nsresult   *res     ///< (out) pointer to where the acquired result code
                        ///<       should be stored, if found
)
{
    Tcl_Obj *errorCode;
    int     objc;
    Tcl_Obj **objv;
    long     longVal;

    NS_PRECONDITION(interp, "null pointer");
    NS_PRECONDITION(res, "null pointer");

    errorCode = Tcl_GetVar2Ex(interp, "errorCode", NULL, TCL_GLOBAL_ONLY);

    if (errorCode && (Tcl_ListObjGetElements(NULL, errorCode, &objc, &objv) == TCL_OK))
    {
        if ((objc == 2) && (!strcmp("XPCOM", Tcl_GetString(objv[0]))))
        {
            if (Tcl_GetLongFromObj(NULL, objv[1], &longVal) == TCL_OK)
            {
                *res = (nsresult)longVal;
                return 1;
            }
        }
    }
    
    return 0;
}


/**
 *  Sets an interpreter's result to an error message corresponding
 *  to the given XPCOM result code.
 */
void
TclXPCOM_SetErrorMessage(
    Tcl_Interp *interp, ///< (in) interpreter to set result
    nsresult    res     ///< (in) XPCOM result code
)
{
    int rv;
    Tcl_CmdInfo cmdinfo;
    Tcl_Obj *script;

    NS_PRECONDITION(interp, "null pointer");

    if (Tcl_GetCommandInfo(interp, TCLXPCOM_ERRORMSGPROC, &cmdinfo))
    {
        script = Tcl_NewListObj(0, NULL);
        Tcl_IncrRefCount(script);
        Tcl_ListObjAppendElement(interp, script, Tcl_NewStringObj(TCLXPCOM_ERRORMSGPROC, -1));
        Tcl_ListObjAppendElement(interp, script, Tcl_NewLongObj((long)res));

        rv = Tcl_EvalObjEx(interp, script, TCL_EVAL_DIRECT | TCL_EVAL_GLOBAL);

        Tcl_DecrRefCount(script);

        if (rv != TCL_OK)
        {
            Tcl_AddObjErrorInfo(interp, "\nwhile executing error message procedure", -1);
        }

    } else
    {
        Tcl_ResetResult(interp);
        Tcl_AppendResult(interp, "xpcom error: unknown (no error message procedure defined)", NULL);
    }
}


/**
 *  Convenience function that sets an interpreter's 'errorCode' variable
 *  and result for a given XPCOM result code.
 */
void
TclXPCOM_SetError(
    Tcl_Interp *interp, ///< (in) interpreter to set result and 'errorCode'
    nsresult    res     ///< (in) XPCOM result code
)
{
    NS_PRECONDITION(interp, "null pointer");

    TclXPCOM_SetErrorMessage(interp, res);
    TclXPCOM_SetErrorCode   (interp, res);
}


/**
 *  Handles background errors caused by Tcl Component bindings. If
 *  the 'xpcom::bgerror' procedure is defined in the given interpreter,
 *  it will be called with the error message in the interpreter's result.
 *  If there is no 'xpcom::bgerror' procedure defined, or an error occurs
 *  while executing the procedure, normal background error processing will
 *  occur.
 *
 *  @note Unlike Tcl_BackgroundError, which invokes 'bgerror' as an idle
 *  event, this function invokes 'xpcom::bgerror' immediately.
 */
void
TclXPCOM_BackgroundError(
    Tcl_Interp *interp  ///< (in) interpreter in which the error occured
)
{
    int rv;
    Tcl_CmdInfo cmdinfo;
    Tcl_Obj *script;

    NS_PRECONDITION(interp, "null pointer");

    if (Tcl_GetCommandInfo(interp, TCLXPCOM_BGERRORPROC, &cmdinfo))
    {
        script = Tcl_NewListObj(0, NULL);
        Tcl_IncrRefCount(script);
        Tcl_ListObjAppendElement(interp, script, Tcl_NewStringObj(TCLXPCOM_BGERRORPROC, -1));
        Tcl_ListObjAppendElement(interp, script, Tcl_GetObjResult(interp));

        rv = Tcl_EvalObjEx(interp, script, TCL_EVAL_DIRECT | TCL_EVAL_GLOBAL);

        Tcl_DecrRefCount(script);

        if (rv != TCL_OK)
        {
            Tcl_BackgroundError(interp);
        }

    } else
    {
        Tcl_BackgroundError(interp);
    }
}
