/* -*- Mode: C++; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is TclXPCOM.
 * 
 * The Initial Developer of the Original Code is Mark Follett.
 * Portions created by Mark Follett are Copyright (C) 2001-2002
 * Mark Follett.  All Rights Reserved.
 * 
 * Contributor(s):
 *     Mark Follett <mef123@myrealbox.com> (Original Author)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 */

/**
 *  @file TclXPCOMInit.cpp
 *  Contains initialization functions for the TclXPCOM library.
 */

#include "TclXPCOMPrivate.h"
#include "nsXPCOM.h"
#include "nsILocalFile.h"
#include "nsIDirectoryService.h"
#include "nsIThread.h"
#include "nsString.h"


nsIServiceManager       *servmanager;///< global reference to the XPCOM Service Manager
nsIInterfaceInfoManager *manager;    ///< global reference to the XPCOM Interface Info Manager
nsIComponentRegistrar   *registrar;  ///< global reference to the XPCOM Component Manager
nsIMemory               *memmgr;     ///< global reference to the XPCOM Memory Allocator

nsIID nsisupports_iid       = NS_ISUPPORTS_IID;
nsIID nsiclassinfo_iid      = NS_ICLASSINFO_IID;
nsIID nsiservicemanager_iid = NS_ISERVICEMANAGER_IID;

nsIInterfaceInfo *nsisupports_info;
nsIInterfaceInfo *nsiservicemanager_info;
nsIInterfaceInfo *nsiclassinfo_info;

Tcl_Encoding ascii_enc;

static int TclXPCOMInitialized = 0;  ///< indicates whether the library has been initialized


/**
 *  Initializes a safe version of TclXPCOM in the given interpreter.
 *  Currently this is identical to the 'non-safe' version.
 *
 *  @retval TCL_OK    - if successful
 *  @retval TCL_ERROR - if initialization failed
 */
int
Tclxpcom_SafeInit(
    Tcl_Interp *interp ///< (in) interpreter to initialize
)
{
    NS_PRECONDITION(interp, "null pointer");

    return Tclxpcom_Init(interp);
}


/**
 *  Initializes the TclXPCOM library if necessary and adds
 *  the 'xpcom' command to the given interpreter. If
 *  '::xpcom::mozilla_home' is set, its value will be used during
 *  the initialization of XPCOM to specify the location of the mozilla
 *  'bin' directory.
 *
 *  @retval TCL_OK    - if successful
 *  @retval TCL_ERROR - if initialization failed
 */
int
Tclxpcom_Init(
    Tcl_Interp *interp ///< (in) interpreter to initialize
)
{
    int rv;
    nsresult res;
    Tcl_Obj *mozhome;
    nsILocalFile *mozBinDirectory;
    nsIDirectoryServiceProvider *appFileLocProvider;
    nsIThread *mainthread;

    NS_PRECONDITION(interp, "null pointer");

    if (Tcl_InitStubs(interp, "8.3", 0) == NULL)
    {
        return TCL_ERROR;
    }

    rv = Tcl_PkgProvide(interp, "tclxpcom", "0.1");
    if (rv != TCL_OK)
    {
        return rv;
    }

    if (!TclXPCOMInitialized)
    {
        ascii_enc = Tcl_GetEncoding(interp, "ascii");
        if (!ascii_enc)
        {
            return TCL_ERROR;
        }

        res = nsIThread::GetMainThread(&mainthread);
        if (NS_FAILED(res))
        {
            mozhome = Tcl_GetVar2Ex(interp, TCLXPCOM_MOZHOMEVARNAME, NULL, TCL_GLOBAL_ONLY);
            if (mozhome)
            {
                res = NS_NewLocalFile(nsString(Tcl_GetUnicode(mozhome), Tcl_GetCharLength(mozhome)), false, &mozBinDirectory);
                if (NS_FAILED(res))
                {
                    Tcl_ResetResult(interp);
                    Tcl_AppendResult(interp, "invalid path \"", Tcl_GetString(mozhome), "\" for mozilla home", NULL);
                    return TCL_ERROR;
                }
            } else
            {
                mozBinDirectory = nsnull;
            }

            appFileLocProvider = nsnull;

            NS_InitXPCOM2(nsnull, mozBinDirectory, appFileLocProvider);

            if (mozhome)
            {
                NS_RELEASE(mozBinDirectory);
            }
        } else
        {
            NS_RELEASE(mainthread);
        }

        res = NS_GetServiceManager(&servmanager);
        if (NS_FAILED(res))
        {
            Tcl_ResetResult(interp);
            Tcl_AppendResult(interp, "unable to get service manager", NULL);
            return TCL_ERROR;
        }

        res = NS_GetComponentRegistrar(&registrar);
        if (NS_FAILED(res))
        {
            Tcl_ResetResult(interp);
            Tcl_AppendResult(interp, "unable to get component registrar", NULL);
            return TCL_ERROR;
        }

        res = NS_GetMemoryManager(&memmgr);
        if (NS_FAILED(res))
        {
            Tcl_ResetResult(interp);
            Tcl_AppendResult(interp, "unable to get memory manager", NULL);
            return TCL_ERROR;
        }

        manager = XPTI_GetInterfaceInfoManager();
        if (!manager)
        {
            Tcl_ResetResult(interp);
            Tcl_AppendResult(interp, "unable to get interfaceinfo manager", NULL);
            return TCL_ERROR;
        }

        res = manager->GetInfoForIID(&nsisupports_iid, &nsisupports_info);
        if (NS_FAILED(res))
        {
            Tcl_ResetResult(interp);
            Tcl_AppendResult(interp, "unable to get nsISupports interfaceinfo", NULL);
            return TCL_ERROR;
        }

        res = manager->GetInfoForIID(&nsiservicemanager_iid, &nsiservicemanager_info);
        if (NS_FAILED(res))
        {
            Tcl_ResetResult(interp);
            Tcl_AppendResult(interp, "unable to get nsIServiceManager interfaceinfo", NULL);
            return TCL_ERROR;
        }

        res = manager->GetInfoForIID(&nsiclassinfo_iid, &nsiclassinfo_info);
        if (NS_FAILED(res))
        {
            Tcl_ResetResult(interp);
            Tcl_AppendResult(interp, "unable to get nsIClassInfo interfaceinfo", NULL);
            return TCL_ERROR;
        }

        TclXPCOM_RegisterObjTypes();
        TclXPCOM_InitInterfaceInfo();
        TclXPCOM_InitClassInfo();
        TclXPCOM_InitInterfaceRef();

        TclXPCOMInitialized = 1;
    }

    TclXPCOM_RegisterCommands(interp);

    return TCL_OK;
}
