/* -*- Mode: C++; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is TclXPCOM.
 * 
 * The Initial Developer of the Original Code is Mark Follett.
 * Portions created by Mark Follett are Copyright (C) 2001-2002
 * Mark Follett.  All Rights Reserved.
 * 
 * Contributor(s):
 *     Mark Follett <mef123@myrealbox.com> (Original Author)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 */

/**
 *  @file TclXPCOMInterfaceUtils.cpp
 *  Contains various utility functions for interface introspection.
 */

#include "TclXPCOMPrivate.h"


/**
 *  Duplicates a chunk of memory using the XPCOM memory
 *  manager.
 *  
 *  @return an allocated copy of the given memory chunk
 */
void *
CloneMemory(
    void   *ptr, ///< (in) memory to clone
    size_t  size ///< (in) size of chunk
)
{
    void *result;

    result = memmgr->Alloc(size);
    memcpy(result, ptr, size);

    return result;
}


/**
 *  Verifies that an interface is reflectable into TclXPCOM (is marked
 *  'scriptable' in the IDL file). If it is not reflectable, an error
 *  message will be set in the given interpreter. If the
 *  'FORCE_NOTSCRIPTABLE' bit is set in 'flags', then no error will
 *  be generated even if the interface is not reflectable.
 *  
 *  @retval TCL_OK    - if the interface is reflectable
 *  @retval TCL_ERROR - if the interface is not reflectable
 */
int
TclXPCOM_CheckInterfaceIsReflectable(
    Tcl_Interp       *interp,       ///< (in) interpreter to store error message
    nsIInterfaceInfo *interfaceinfo,///< (in) interface to verify
    int               flags         ///< (in) override flags
)
{
    PRBool scriptable;
    const char *name;

    NS_PRECONDITION(interp, "null pointer");
    NS_PRECONDITION(interfaceinfo, "null pointer");
    NS_PRECONDITION(!(flags & ~(FORCE_HIDDEN | FORCE_NOTXPCOM | FORCE_NOTSCRIPTABLE)), "invalid flags");

    if (!(flags & FORCE_NOTSCRIPTABLE))
    {
        interfaceinfo->IsScriptable(&scriptable);
        if (!scriptable)
        {
            interfaceinfo->GetNameShared(&name);
            Tcl_ResetResult(interp);
            Tcl_AppendResult(interp, "interface \"", name, "\" is not scriptable", NULL);
            return TCL_ERROR;
        }
    }
    return TCL_OK;
}


/**
 *  Verifies that a method is reflectable into TclXPCOM (is not marked
 *  'noscript' or 'notxpcom' in the IDL file). If it is not reflectable,
 *  an error message will be set in the given interpreter. The
 *  'FORCE_HIDDEN' and 'FORCE_NOTXPCOM' bits in 'flags' will override
 *  any 'noscript' or 'notxpcom' marked methods, respectively, forcing
 *  a successful return.
 *  
 *  @retval TCL_OK    - if the method is reflectable
 *  @retval TCL_ERROR - if the method is not reflectable
 */
int
TclXPCOM_CheckMethodIsReflectable(
    Tcl_Interp       *interp,       ///< (in) interpreter to store error message
    nsIInterfaceInfo *interfaceinfo,///< (in) interface of method to verify
    PRUint16          methodindex,  ///< (in) index into interface of method to verify
    int               flags         ///< (in) override flags
)
{
    const char *name;
    const nsXPTMethodInfo *methodinfo;

    NS_PRECONDITION(interp, "null pointer");
    NS_PRECONDITION(interfaceinfo, "null pointer");
    NS_PRECONDITION(!(flags & ~(FORCE_HIDDEN | FORCE_NOTXPCOM | FORCE_NOTSCRIPTABLE)), "invalid flags");

    interfaceinfo->GetMethodInfo(methodindex, &methodinfo);

    if (methodinfo->IsNotXPCOM() && !(flags & FORCE_NOTXPCOM))
    {
        name = methodinfo->GetName();
        Tcl_ResetResult(interp);
        Tcl_AppendResult(interp, "\"", name, "\" is not an XPCOM method", NULL);
        return TCL_ERROR;
    }

    if (!(flags & FORCE_HIDDEN) && methodinfo->IsHidden())
    {
        name = methodinfo->GetName();
        Tcl_ResetResult(interp);
        Tcl_AppendResult(interp, "\"", name, "\" is a hidden method", NULL);
        return TCL_ERROR;
    }

    return TCL_OK;
}


/**
 *  Determines whether an interface specified by 'iid' is an
 *  ancestor of the interface specified by 'interfaceinfo', or
 *  if 'iid' and 'interfaceinfo' refer to the same interface.
 *  
 *  @retval 1 - if 'iid' is equal to or an ancestor of 'interfaceinfo'
 *  @retval 0 - if 'iid' is not in 'interfaceinfo'
 */
int
isInInterface(
    nsIInterfaceInfo *interfaceinfo,///< (in) interfaceinfo of interface
    nsIID            *iid           ///< (in) iid of interface to check
)
{
    PRBool isiid;
    PRBool hasancestor;

    NS_PRECONDITION(interfaceinfo, "null pointer");
    NS_PRECONDITION(iid, "null pointer");

    interfaceinfo->IsIID(iid, &isiid);
    interfaceinfo->HasAncestor(iid, &hasancestor);

    if (isiid || hasancestor)
    {
        return 1;
    }

    return 0;
}


/**
 *  Lists the constants of an interface and its ancestors.
 *
 *  @return a Tcl list of constant names
 */
Tcl_Obj *
TclXPCOM_ListInterfaceConstants(
    nsIInterfaceInfo *interfaceinfo ///< (in) interface to search
)
{
    PRUint16 i;
    Tcl_Obj *result;
    InterfaceInfo *iinfo;

    NS_PRECONDITION(interfaceinfo, "null pointer");

    iinfo = GetInterfaceInfo(interfaceinfo);

    result = Tcl_NewListObj(0, NULL);
    for (i = 0; i < iinfo->constantcount; i++)
    {
        Tcl_ListObjAppendElement(NULL, result, Tcl_NewStringObj(iinfo->constantnames[i], -1));
    }

    return result;
}


/**
 *  Lists the attributes of an interface and its ancestors.
 *
 *  @return a Tcl list of attribute names
 */
Tcl_Obj *
TclXPCOM_ListInterfaceAttributes(
    nsIInterfaceInfo *interfaceinfo ///< (in) interface to search
)
{
    PRUint16 i;
    Tcl_Obj *result;
    InterfaceInfo *iinfo;

    NS_PRECONDITION(interfaceinfo, "null pointer");

    iinfo = GetInterfaceInfo(interfaceinfo);

    result = Tcl_NewListObj(0, NULL);
    for (i = 0; i < iinfo->attributecount; i++)
    {
        Tcl_ListObjAppendElement(NULL, result, Tcl_NewStringObj(iinfo->attributes[i].name, -1));
    }

    return result;
}


/**
 *  Lists the methods of an interface and its ancestors.
 *
 *  @return a Tcl list of method names
 */
Tcl_Obj *
TclXPCOM_ListInterfaceMethods(
    nsIInterfaceInfo *interfaceinfo ///< (in) interface to search
)
{
    PRUint16 i;
    Tcl_Obj *result;
    InterfaceInfo *iinfo;

    NS_PRECONDITION(interfaceinfo, "null pointer");

    iinfo = GetInterfaceInfo(interfaceinfo);

    result = Tcl_NewListObj(0, NULL);
    for (i = 0; i < iinfo->methodcount; i++)
    {
        Tcl_ListObjAppendElement(NULL, result, Tcl_NewStringObj(iinfo->methods[i].name, -1));
    }

    return result;
}


/**
 *  Lists the methods associated with a set of interfaces composing
 *  a class. Any ambiguous method names will be qualified with their
 *  corresponding interface name.
 *
 *  @return a Tcl list of method names
 */
Tcl_Obj *
TclXPCOM_ListClassMethods(
    ClassInfo *cinfo ///< (in) class to search
)
{
    PRUint16 i;
    Tcl_Obj *result;

    NS_PRECONDITION(cinfo, "null pointer");

    result = Tcl_NewListObj(0, NULL);
    for (i = 0; i < cinfo->methodcount; i++)
    {
        Tcl_ListObjAppendElement(NULL, result, Tcl_NewStringObj(cinfo->methods[i].idinfo.name, -1));
    }

    return result;
}


/**
 *  Lists the attributes associated with a set of interfaces composing
 *  a class. Any ambiguous attribute names will be qualified with their
 *  corresponding interface name.
 *
 *  @return a Tcl list of attribute names
 */
Tcl_Obj *
TclXPCOM_ListClassAttributes(
    ClassInfo *cinfo ///< (in) class to search
)
{
    PRUint16 i;
    Tcl_Obj *result;

    NS_PRECONDITION(cinfo, "null pointer");

    result = Tcl_NewListObj(0, NULL);
    for (i = 0; i < cinfo->attributecount; i++)
    {
        Tcl_ListObjAppendElement(NULL, result, Tcl_NewStringObj(cinfo->attributes[i].idinfo.name, -1));
    }

    return result;
}


/**
 *  Determines the string representation of the type of a given
 *  parameter. If the parameter is an array, a list will be returned
 *  whose first element is 'list' followed by the type of values
 *  in the array.
 *
 *  @return a new Tcl object with refcount 0 holding the string
 *          representation of the method parameter's type.
 */
Tcl_Obj *
GetType(
    nsIInterfaceInfo *interfaceinfo,///< (in) interface to query
    PRUint16          methodindex,  ///< (in) index of method
    nsXPTParamInfo   *paraminfo,    ///< (in) xpt parameter info structure of parameter
    nsXPTType         type          ///< (in) xpt type structure of parameter
)
{
    nsXPTType arraytype;
    nsIInterfaceInfo *info;
    const char *name;
    Tcl_Obj *result;
    nsresult res;
    const nsXPTMethodInfo *methodinfo;
    char str[128];

    NS_PRECONDITION(interfaceinfo, "null pointer");
    NS_PRECONDITION(paraminfo, "null pointer");
    // type checked in default case

    switch (type.TagPart())
    {
        case nsXPTType::T_I8:
            result = Tcl_NewStringObj("int8", -1);
            break;
        case nsXPTType::T_I16:
            result = Tcl_NewStringObj("int16", -1);
            break;
        case nsXPTType::T_I32:
            result = Tcl_NewStringObj("int32", -1);
            break;
        case nsXPTType::T_I64:
            result = Tcl_NewStringObj("int64", -1);
            break;
        case nsXPTType::T_U8:
            result = Tcl_NewStringObj("uint8", -1);
            break;
        case nsXPTType::T_U16:
            result = Tcl_NewStringObj("uint16", -1);
            break;
        case nsXPTType::T_U32:
            result = Tcl_NewStringObj("uint32", -1);
            break;
        case nsXPTType::T_U64:
            result = Tcl_NewStringObj("uint64", -1);
            break;
        case nsXPTType::T_FLOAT:
            result = Tcl_NewStringObj("float", -1);
            break;
        case nsXPTType::T_DOUBLE:
            result = Tcl_NewStringObj("double", -1);
            break;
        case nsXPTType::T_BOOL:
            result = Tcl_NewStringObj("boolean", -1);
            break;
        case nsXPTType::T_CHAR:
            result = Tcl_NewStringObj("char", -1);
            break;
        case nsXPTType::T_WCHAR:
            result = Tcl_NewStringObj("wchar", -1);
            break;
        case nsXPTType::T_VOID:
            if (type.IsPointer())
            {
                result = Tcl_NewStringObj("pointer", -1);
            } else
            {
                result = Tcl_NewStringObj("void", -1);
            }
            break;
        case nsXPTType::T_IID:
            result = Tcl_NewStringObj("interface", -1);
            break;

        case nsXPTType::T_ASTRING:
        case nsXPTType::T_DOMSTRING:
        case nsXPTType::T_UTF8STRING:
        case nsXPTType::T_WCHAR_STR:
        case nsXPTType::T_PWSTRING_SIZE_IS:
            result = Tcl_NewStringObj("wstring", -1);
            break;

        case nsXPTType::T_CSTRING:
        case nsXPTType::T_CHAR_STR:
        case nsXPTType::T_PSTRING_SIZE_IS:
            result = Tcl_NewStringObj("string", -1);
            break;

        case nsXPTType::T_INTERFACE:
            res = interfaceinfo->GetInfoForParam(methodindex, paraminfo, &info);
            if (NS_SUCCEEDED(res))
            {
                info->GetNameShared(&name);
                result = Tcl_NewStringObj(name, -1);
                NS_RELEASE(info);
            } else {
                interfaceinfo->GetMethodInfo(methodindex, &methodinfo);
                interfaceinfo->GetNameShared(&name);
                sprintf(str, "GetInfoForParam Failed: Interface %s Method %s", name, methodinfo->GetName());
                NS_WARNING(str);

                result = Tcl_NewStringObj("nsISupports", -1);
            }
            break;

        case nsXPTType::T_INTERFACE_IS:
            result = Tcl_NewStringObj("nsISupports", -1);
            break;

        case nsXPTType::T_ARRAY:
            interfaceinfo->GetTypeForParam(methodindex, paraminfo, 1, &arraytype);
            result = Tcl_NewListObj(0, NULL);
            Tcl_ListObjAppendElement(NULL, result, Tcl_NewStringObj("list", -1));
            Tcl_ListObjAppendElement(NULL, result, GetType(interfaceinfo, methodindex, paraminfo, arraytype));
            break;

        default:
            NS_ERROR("invalid type");
            break;
    }

    return result;
}


/**
 *  Determines the string representation of a method parameter.
 *  It's representation consists of a list whose first element is
 *  the direction of the parameter (in, out, or inout) followed
 *  by the type of the parameter.
 *
 *  @return a new Tcl object with refcount 0 holding the string
 *          representation of the method parameter.
 */
Tcl_Obj *
GetParam(
    nsIInterfaceInfo *interfaceinfo,///< (in) interface information of method
    PRUint16          methodindex,  ///< (in) index into interface for method
    nsXPTParamInfo   *paraminfo     ///< (in) xpt parameter info structure of parameter
)
{
    Tcl_Obj *result, *direction;

    NS_PRECONDITION(interfaceinfo, "null pointer");
    NS_PRECONDITION(paraminfo, "null pointer");

    result = Tcl_NewListObj(0, NULL);

    if (paraminfo->IsIn() && !paraminfo->IsDipper())
    {
        if (paraminfo->IsOut())
        {
            direction = Tcl_NewStringObj("inout", -1);
        } else
        {
            direction = Tcl_NewStringObj("in", -1);
        }
    } else
    {
        direction = Tcl_NewStringObj("out", -1);
    }

    Tcl_ListObjAppendElement(NULL, result, direction);
    Tcl_ListObjAppendElement(NULL, result, GetType(interfaceinfo, methodindex, paraminfo, paraminfo->GetType()));

    return result;
}


/**
 *  Lists the types of a method's arguments.
 *  
 *  @return a new Tcl object with refcount 0 holding a list
 *          of the method's argument types.
 */
Tcl_Obj *
TclXPCOM_ListMethodArguments(
    nsIInterfaceInfo *interfaceinfo,///< (in) interface information for method
    PRUint16          methodindex   ///< (in) index into interface for method
)
{
    PRUint8 i;
    Tcl_Obj *result;
    MethodInfo *minfo;
    PRUint16 *tclparams;

    NS_PRECONDITION(interfaceinfo, "null pointer");

    minfo = GetMethodInfo(interfaceinfo, methodindex);
    tclparams = minfo->tclparams;

    result = Tcl_NewListObj(0, NULL);

    for (i = 0; i < minfo->paramcount; i++)
    {
        if ((tclparams[i] & PARAM_IS_SKIPPED) ||
            (tclparams[i] & PARAM_IS_RETVAL))
        {
            continue;
        }

        Tcl_ListObjAppendElement(NULL, result, GetParam(interfaceinfo, methodindex, &minfo->paraminfos[i]));
    }

    return result;
}


/**
 *  Determines the type of a method's return value.
 *
 *  @return a new Tcl object with refcount 0 holding the string
 *          representation of the method's return value
 */
Tcl_Obj *
TclXPCOM_GetReturnValueType(
    nsIInterfaceInfo *interfaceinfo,///< (in) interface information for method
    IdentifierInfo   *idinfo        ///< (in) identifier information for method
)
{
    PRUint8 i;
    PRUint16 methodindex;
    MethodInfo *minfo;
    PRUint16 *tclparams;

    NS_PRECONDITION(interfaceinfo, "null pointer");
    NS_PRECONDITION(idinfo, "null pointer");

    methodindex = idinfo->indices.methodindex;

    minfo = GetMethodInfo(interfaceinfo, methodindex);
    tclparams = minfo->tclparams;

    for (i = 0; i < minfo->paramcount; i++)
    {
        if (tclparams[i] & PARAM_IS_RETVAL)
        {
            return GetType(interfaceinfo, methodindex, &minfo->paraminfos[i], minfo->paraminfos[i].GetType());
        }
    }

    return Tcl_NewObj();
}


/**
 *  Determines the type of an attribute.
 *
 *  @return a new Tcl object with refcount 0 holding the string
 *          representation of the attribute's type
 */
Tcl_Obj *
TclXPCOM_GetAttributeType(
    nsIInterfaceInfo *interfaceinfo,///< (in) interface information for the attribute
    IdentifierInfo   *idinfo        ///< (in) indentifier information for the attribute
)
{
    PRUint16 methodindex;
    MethodInfo *minfo;
    PRUint16 *tclparams;

    NS_PRECONDITION(interfaceinfo, "null pointer");
    NS_PRECONDITION(idinfo, "null pointer");

    if (idinfo->type & INTERFACEINFO_GETTER)
    {
        methodindex = idinfo->indices.attributeindices.getterindex;
    } else
    {
        methodindex = idinfo->indices.attributeindices.setterindex;
    }

    minfo = GetMethodInfo(interfaceinfo, methodindex);
    tclparams = minfo->tclparams;

    return GetType(interfaceinfo, methodindex, &minfo->paraminfos[0], minfo->paraminfos[0].GetType());
}
