/* -*- Mode: C++; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is TclXPCOM.
 * 
 * The Initial Developer of the Original Code is Mark Follett.
 * Portions created by Mark Follett are Copyright (C) 2001-2002
 * Mark Follett.  All Rights Reserved.
 * 
 * Contributor(s):
 *     Mark Follett <mef123@myrealbox.com> (Original Author)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 */

/**
 *  @file TclXPCOMMarshall.cpp
 *  Functions for converting between Tcl and XPCOM values.
 *
 *  @see http://www.mozilla.org/scriptable/typelib_file.html#TypeDescriptor
 */

#include "TclXPCOMPrivate.h"
#include "nsAString.h"
#include "nsReadableUtils.h"

// XXX TODO: use nsAUTF8String when it appears

/**
 *  Determines the size of an XPT type.
 *
 *  @return the size in bytes of the given type, or 0 if
 *          the typecode is invalid
 */
size_t
GetElementTypeSize(
    nsXPTType *type ///< (in) an XPT typecode
)
{
    size_t size;

    // type checked in default case

    if (type->IsPointer())
    {
        size = sizeof(void*);
    } else
    {
        switch (type->TagPart())
        {
            case nsXPTType::T_I8:
                size = sizeof(PRInt8);
                break;
            case nsXPTType::T_I16:
                size = sizeof(PRInt16);
                break;
            case nsXPTType::T_I32:
                size = sizeof(PRInt32);
                break;
            case nsXPTType::T_I64:
                size = sizeof(PRInt64);
                break;
            case nsXPTType::T_U8:
                size = sizeof(PRUint8);
                break;
            case nsXPTType::T_U16:
                size = sizeof(PRUint16);
                break;
            case nsXPTType::T_U32:
                size = sizeof(PRUint32);
                break;
            case nsXPTType::T_U64:
                size = sizeof(PRUint64);
                break;
            case nsXPTType::T_FLOAT:
                size = sizeof(float);
                break;
            case nsXPTType::T_DOUBLE:
                size = sizeof(double);
                break;
            case nsXPTType::T_BOOL:
                size = sizeof(PRBool);
                break;
            case nsXPTType::T_CHAR:
                size = sizeof(char);
                break;
            case nsXPTType::T_WCHAR:
                size = sizeof(PRUnichar);
                break;

            case nsXPTType::T_VOID: // ?
            default:
                NS_ERROR("invalid type");
                size = 0;  // invalid
                break;
        }
    }
    return size;
}


/**
 *  Frees a parameter, if appropriate for the given type.
 *
 *  @note Void* parameters are not freed: calling non-scriptable methods with
 *        these may cause memory leaks.
 */
void
CleanupParam(
    void             **params,       ///< (in) array of parameters in method
    void              *param,        ///< (in) address of parameter to free
    nsXPTType         *type,         ///< (in) XPT typecode of the parameter to free
    nsIInterfaceInfo  *interfaceinfo,///< (in) interface information for the method
    PRUint16           methodindex,  ///< (in) index into interface of method
    nsXPTParamInfo    *paraminfo     ///< (in) array of parameter info for method
)
{
    PRUint32 i, j;
    size_t elementsize;
    PRUint8 lengthindex;
    nsXPTType arraytype;
    nsISupports *isupports;

    NS_PRECONDITION(params, "null pointer");
    NS_PRECONDITION(param, "null pointer");
    NS_PRECONDITION(type, "null pointer");
    NS_PRECONDITION(interfaceinfo, "null pointer");
    NS_PRECONDITION(paraminfo, "null pointer");

    if (type->IsPointer() && (*(void **)param != NULL))
    {
        switch (type->TagPart())
        {
            case nsXPTType::T_VOID:
                // don't free a void pointer
                break;

            case nsXPTType::T_ASTRING:
            case nsXPTType::T_DOMSTRING:
                delete *(nsAString **)param;
                break;

            case nsXPTType::T_CSTRING:
            case nsXPTType::T_UTF8STRING:
                delete *(nsACString **)param;
                break;

            case nsXPTType::T_INTERFACE:
            case nsXPTType::T_INTERFACE_IS:
                isupports = *(nsISupports **)param;
                NS_RELEASE(isupports);
                break;

            case nsXPTType::T_ARRAY:
                interfaceinfo->GetTypeForParam(methodindex, paraminfo, 1, &arraytype);
                interfaceinfo->GetLengthIsArgNumberForParam(methodindex, paraminfo, 0, &lengthindex);
                elementsize = GetElementTypeSize(&arraytype);
                j = *(PRUint32 *)params[lengthindex];
                for (i = 0; i < j; i++)
                {
                    CleanupParam(params, (void *)(*(char **)param + (i * elementsize)), &arraytype, interfaceinfo, methodindex, paraminfo);
                }
                memmgr->Free(*(void **)param);
                break;

            default:
                if (!paraminfo->IsShared())
                {
                    memmgr->Free(*(void **)param);
                }
                break;
        }
    }
}


/**
 *  Converts a native value to a new Tcl object whose refcount is 0.
 *  
 *  @retval NS_OK                 - if successful
 *  @retval NS_ERROR_INVALID_ARG  - if the given typecode is invalid
 *  @retval NS_ERROR_NULL_POINTER - if the parameter cannot be a null pointer
 */
nsresult
Native2Tcl(
    void             **params,       ///< (in)  array of parameters in method
    void              *param,        ///< (in)  address of parameter to convert
    nsXPTType         *type,         ///< (in)  XPT typecode of the parameter to convert
    nsIInterfaceInfo  *interfaceinfo,///< (in)  interface information for the method
    PRUint16           methodindex,  ///< (in)  index into interface of method
    nsXPTParamInfo    *paraminfo,    ///< (in)  array of parameter info for method
    Tcl_Obj          **result        ///< (out) address to store a reference to the created Tcl Object
)
{
    nsresult res;
    PRUint32 i, j;
    PRUint8 lengthindex, iidindex;
    size_t elementsize;
    Tcl_Obj *listitem;
    nsXPTType arraytype;
    nsIID *iid;
    nsAString *astr;
    nsACString *acstr;
    char *charPtr;
    PRUnichar *wcharPtr;
    nsIInterfaceInfo *tempinfo;

    NS_PRECONDITION(params, "null pointer");
    NS_PRECONDITION(param, "null pointer");
    NS_PRECONDITION(type, "null pointer");
//  PrepareInterfaceConstantValue calls function with interfaceinfo as NULL
//  NS_PRECONDITION(interfaceinfo, "null pointer");
//  PrepareInterfaceConstantValue calls function with paraminfo as NULL
//  NS_PRECONDITION(paraminfo, "null pointer");
    NS_PRECONDITION(result, "null pointer");

    if (!(type->IsPointer()))
    {
        switch (type->TagPart())
        {
            case nsXPTType::T_I8:
                *result = Tcl_NewIntObj((int)*(PRInt8 *)param);
                break;

            case nsXPTType::T_I16:
                *result = Tcl_NewIntObj((int)*(PRInt16 *)param);
                break;

            case nsXPTType::T_I32:
                *result = Tcl_NewLongObj((long)*(PRInt32 *)param);
                break;

            case nsXPTType::T_I64:
                *result = TclXPCOM_NewWideIntObj((TclXPCOM_WideInt)*(PRInt64 *)param);
                break;

            case nsXPTType::T_U8:
                *result = Tcl_NewIntObj((int)*(PRUint8 *)param);
                break;

            case nsXPTType::T_U16:
                *result = Tcl_NewIntObj((int)*(PRUint16 *)param);
                break;

            case nsXPTType::T_U32:
                *result = Tcl_NewLongObj((long)*(PRUint32 *)param);
                break;

            case nsXPTType::T_U64:
                *result = TclXPCOM_NewWideIntObj((TclXPCOM_WideInt)*(PRUint64 *)param);
                break;

            case nsXPTType::T_FLOAT:
                *result = Tcl_NewDoubleObj((double)*(float *)param);
                break;

            case nsXPTType::T_DOUBLE:
                *result = Tcl_NewDoubleObj(*(double *)param);
                break;

            case nsXPTType::T_BOOL:
                *result = Tcl_NewBooleanObj((int)*(PRBool *)param);
                break;

            case nsXPTType::T_CHAR:
                *result = Tcl_NewStringObj((char *)param, 1);
                break;

            case nsXPTType::T_WCHAR:
                *result = Tcl_NewUnicodeObj((Tcl_UniChar *)param, 1);
                break;

            case nsXPTType::T_VOID:
                *result = Tcl_NewStringObj("", -1);
                break;

            default:
                // xpcom error,  invalid type
                *result = NULL;
                return NS_ERROR_INVALID_ARG;
                break;
        }
    } else
    {
        if (*(void **)param == NULL)
        {
            if (type->IsReference())
            {
                // xpcom error, calling function is crappy
                *result = NULL;
                return NS_ERROR_NULL_POINTER;
            } else
            {
                *result = TclXPCOM_NewPointerObj(NULL);
                return NS_OK;
            }
        }

        switch (type->TagPart())
        {
            case nsXPTType::T_ASTRING:
            case nsXPTType::T_DOMSTRING:
                if ((*(nsAString **)param)->IsVoid())
                {
                    *result = TclXPCOM_NewPointerObj(NULL);
                    return NS_OK;
                }
                break;

            case nsXPTType::T_CSTRING:
            case nsXPTType::T_UTF8STRING:
                if ((*(nsACString **)param)->IsVoid())
                {
                    *result = TclXPCOM_NewPointerObj(NULL);
                    return NS_OK;
                }
                break;

            default:
                break;
        }

        switch (type->TagPart())
        {
            case nsXPTType::T_I8:
                *result = Tcl_NewIntObj((int)**(PRInt8 **)param);
                break;

            case nsXPTType::T_I16:
                *result = Tcl_NewIntObj((int)**(PRInt16 **)param);
                break;

            case nsXPTType::T_I32:
                *result = Tcl_NewLongObj((long)**(PRInt32 **)param);
                break;

            case nsXPTType::T_I64:
                *result = TclXPCOM_NewWideIntObj((TclXPCOM_WideInt)**(PRInt64 **)param);
                break;

            case nsXPTType::T_U8:
                *result = Tcl_NewIntObj((int)**(PRUint8 **)param);
                break;

            case nsXPTType::T_U16:
                *result = Tcl_NewIntObj((int)**(PRUint16 **)param);
                break;

            case nsXPTType::T_U32:
                *result = Tcl_NewLongObj((long)**(PRUint32 **)param);
                break;

            case nsXPTType::T_U64:
                *result = TclXPCOM_NewWideIntObj((TclXPCOM_WideInt)**(PRUint64 **)param);
                break;

            case nsXPTType::T_FLOAT:
                *result = Tcl_NewDoubleObj((double)**(float **)param);
                break;

            case nsXPTType::T_DOUBLE:
                *result = Tcl_NewDoubleObj(**(double **)param);
                break;

            case nsXPTType::T_BOOL:
                *result = Tcl_NewBooleanObj((int)**(PRBool **)param);
                break;

            case nsXPTType::T_CHAR:
                *result = Tcl_NewStringObj(*(char **)param, 1);
                break;

            case nsXPTType::T_WCHAR:
                *result = Tcl_NewUnicodeObj(*(Tcl_UniChar **)param, 1);
                break;

            case nsXPTType::T_VOID:
                *result = TclXPCOM_NewPointerObj(*(void **)param);
                break;

            case nsXPTType::T_IID:
                *result = TclXPCOM_NewNsIDObj(*(nsID **)param);
                break;

            case nsXPTType::T_ASTRING:
            case nsXPTType::T_DOMSTRING:
                // XXX TODO: write directly into the Tcl object's unicode buffer
                astr = *(nsAString **)param;
                wcharPtr = ToNewUnicode(*astr);
                *result = Tcl_NewUnicodeObj((Tcl_UniChar *)wcharPtr, astr->Length());
                memmgr->Free(wcharPtr);
                break;

            case nsXPTType::T_CSTRING:
            case nsXPTType::T_UTF8STRING:
                // XXX TODO: write directly into the Tcl object's buffer
                acstr = *(nsACString **)param;
                charPtr = ToNewCString(*acstr);
                *result = Tcl_NewStringObj(charPtr, acstr->Length());
                memmgr->Free(charPtr);
                break;

            case nsXPTType::T_CHAR_STR:
                *result = Tcl_NewStringObj(*(char **)param, -1);
                break;

            case nsXPTType::T_WCHAR_STR:
                *result = Tcl_NewUnicodeObj(*(Tcl_UniChar **)param, -1);
                break;

            case nsXPTType::T_INTERFACE:
                res = interfaceinfo->GetInfoForParam(methodindex, paraminfo, &tempinfo);
                NS_ASSERTION((NS_SUCCEEDED(res)), "GetInfoForParam failed");

                *result = TclXPCOM_NewISupportsObj(*(nsISupports **)param, tempinfo);
                NS_RELEASE(tempinfo);
                break;

            case nsXPTType::T_INTERFACE_IS:
                interfaceinfo->GetInterfaceIsArgNumberForParam(methodindex, paraminfo, &iidindex);
                iid = *(nsIID **)params[iidindex];

                // XXX will this always return a valid interface ref?
                res = manager->GetInfoForIID(iid, &tempinfo);
                NS_ASSERTION((NS_SUCCEEDED(res)), "GetInfoForIID failed");

                *result = TclXPCOM_NewISupportsObj(*(nsISupports **)param, tempinfo);
                NS_RELEASE(tempinfo);
                break;

            case nsXPTType::T_ARRAY:
                interfaceinfo->GetTypeForParam(methodindex, paraminfo, 1, &arraytype);
                elementsize = GetElementTypeSize(&arraytype);
                interfaceinfo->GetLengthIsArgNumberForParam(methodindex, paraminfo, 0, &lengthindex);
                j = *(PRUint32 *)params[lengthindex];
                *result = Tcl_NewListObj(0, NULL);
                for (i = 0; i < j; i++)
                {
                    res = Native2Tcl(params, (void *)(*(char **)param + (i * elementsize)), &arraytype, interfaceinfo, methodindex, paraminfo, &listitem);
                    if (NS_FAILED(res))
                    {
                        Tcl_DecrRefCount(*result);
                        *result = NULL;
                        return res;
                    }
                    
                    Tcl_ListObjAppendElement(NULL, *result, listitem);
                }
                break;

            case nsXPTType::T_PSTRING_SIZE_IS:
                interfaceinfo->GetLengthIsArgNumberForParam(methodindex, paraminfo, 0, &lengthindex);
                *result = Tcl_NewStringObj(*(char **)param, (int)*(PRUint32 *)params[lengthindex]);
                break;

            case nsXPTType::T_PWSTRING_SIZE_IS:
                interfaceinfo->GetLengthIsArgNumberForParam(methodindex, paraminfo, 0, &lengthindex);
                *result = Tcl_NewUnicodeObj(*(Tcl_UniChar **)param, (int)*(PRUint32 *)params[lengthindex]);
                break;

            default:
                // xpcom error,  invalid type
                *result = NULL;
                return NS_ERROR_INVALID_ARG;
                break;
        }
    }

    return NS_OK;
}


/**
 *  Converts a Tcl Object to a native value. If an error occurs,
 *  an error message will be stored in the interpreter's result
 *  and the 'errorCode' variable will be set to the corresponding
 *  XPCOM return code.
 *  
 *  @retval TCL_OK    - if successful
 *  @retval TCL_ERROR - if an error occurs during conversion
 */
int
Tcl2Native(
    Tcl_Interp       *interp,       ///< (in)  interpreter to hold error message
    Tcl_Obj          *value,        ///< (in)  Tcl Object to convert
    void            **params,       ///< (in)  array of parameters in method
    void             *param,        ///< (out) address to store converted parameter
    nsXPTType        *type,         ///< (in)  XPT typecode of parameter
    int               shared,       ///< (in)  flag indicating parameter is shared & dependent
    nsIInterfaceInfo *interfaceinfo,///< (in)  interface information for the method
    PRUint16          methodindex,  ///< (in)  index into interface of method
    nsXPTParamInfo   *paraminfo     ///< (in)  array of parameter info for method
)
{
    int rv;
    nsresult res;
    PRUint32 i, length;
    PRUint8 sizeindex, lengthindex, iidindex;
    size_t elementsize;
    int        intVal;
    long       longVal;
    double     doubleVal;
    TclXPCOM_WideInt wideVal;
    nsIID     *iidPtr;
    char      *charPtr, *utf8str;
    Tcl_Obj  **listPtr;
    char str[128];
    nsXPTType arraytype;
    nsIInterfaceInfo *tempinfo;
    nsISupports      *isupports;
    int numbytes;

    NS_PRECONDITION(interp, "null pointer");
    NS_PRECONDITION(value, "null pointer");
    NS_PRECONDITION(params, "null pointer");
    NS_PRECONDITION(param, "null pointer");
    NS_PRECONDITION(type, "null pointer");
    // no precondition on shared
    NS_PRECONDITION(interfaceinfo, "null pointer");
    NS_PRECONDITION(paraminfo, "null pointer");

    if (!(type->IsPointer()))
    {
        switch (type->TagPart())
        {
            case nsXPTType::T_I8:
                rv = Tcl_GetIntFromObj(interp, value, &intVal);
                if (rv == TCL_OK)
                {
                    *(PRInt8 *)param = (PRInt8)intVal;
                }
                break;

            case nsXPTType::T_I16:
                rv = Tcl_GetIntFromObj(interp, value, &intVal);
                if (rv == TCL_OK)
                {
                    *(PRInt16 *)param = (PRInt16)intVal;
                }
                break;

            case nsXPTType::T_I32:
                rv = Tcl_GetLongFromObj(interp, value, &longVal);
                if (rv == TCL_OK)
                {
                    *(PRInt32 *)param = (PRInt32)longVal;
                }
                break;

            case nsXPTType::T_I64:
                rv = TclXPCOM_GetWideIntFromObj(interp, value, &wideVal);
                if (rv == TCL_OK)
                {
                    *(PRInt64 *)param = (PRInt64)wideVal;
                }
                break;

            case nsXPTType::T_U8:
                rv = Tcl_GetIntFromObj(interp, value, &intVal);
                if (rv == TCL_OK)
                {
                    *(PRUint8 *)param = (PRUint8)intVal;
                }
                break;

            case nsXPTType::T_U16:
                rv = Tcl_GetIntFromObj(interp, value, &intVal);
                if (rv == TCL_OK)
                {
                    *(PRUint16 *)param = (PRUint16)intVal;
                }
                break;

            case nsXPTType::T_U32:
                rv = Tcl_GetLongFromObj(interp, value, &longVal);
                if (rv == TCL_OK)
                {
                    *(PRUint32 *)param = (PRUint32)longVal;
                }
                break;

            case nsXPTType::T_U64:
                rv = TclXPCOM_GetWideIntFromObj(interp, value, &wideVal);
                if (rv == TCL_OK)
                {
                    *(PRUint64 *)param = (PRUint64)wideVal;
                }
                break;

            case nsXPTType::T_FLOAT:
                rv = Tcl_GetDoubleFromObj(interp, value, &doubleVal);
                if (rv == TCL_OK)
                {
                    *(float *)param = (float)doubleVal;
                }
                break;

            case nsXPTType::T_DOUBLE:
                rv = Tcl_GetDoubleFromObj(interp, value, &doubleVal);
                if (rv == TCL_OK)
                {
                    *(double *)param = doubleVal;
                }
                break;

            case nsXPTType::T_BOOL:
                rv = Tcl_GetBooleanFromObj(interp, value, &intVal);
                if (rv == TCL_OK)
                {
                    *(PRBool *)param = (PRBool)intVal;
                }
                break;

            case nsXPTType::T_CHAR:
                if (Tcl_GetCharLength(value) != 1)
                {
                    Tcl_ResetResult(interp);
                    Tcl_AppendResult(interp, "parameter must be a single character", NULL);
                    rv = TCL_ERROR;
                } else
                {
                    utf8str = Tcl_GetStringFromObj(value, &numbytes);
                    rv = Tcl_UtfToExternal(interp, ascii_enc, utf8str, numbytes, 0, NULL, str, 2, NULL, NULL, NULL);
                    if (rv == TCL_OK)
                    {
                        *(char *)param = str[0];
                    }
                }
                break;

            case nsXPTType::T_WCHAR:
                if (Tcl_GetCharLength(value) != 1)
                {
                    Tcl_ResetResult(interp);
                    Tcl_AppendResult(interp, "parameter must be a single character", NULL);
                    rv = TCL_ERROR;
                } else
                {
                    *(PRUnichar *)param = (PRUnichar)*Tcl_GetUnicode(value);
                    rv = TCL_OK;
                }
                break;

            case nsXPTType::T_VOID:
                // do nothing
                rv = TCL_OK;
                break;

            default:
                // xpcom error,  invalid type
                TclXPCOM_SetError(interp, NS_ERROR_INVALID_ARG);
                rv = TCL_ERROR;
                break;
        }       
    } else
    {
        if (TclXPCOM_IsNULL(value))
        {
            switch (type->TagPart())
            {
                case nsXPTType::T_ASTRING:
                case nsXPTType::T_DOMSTRING:
                    (*(nsAString **)param)->Truncate();
                    (*(nsAString **)param)->SetIsVoid(PR_TRUE);
                    rv = TCL_OK;
                    break;

                case nsXPTType::T_CSTRING:
                case nsXPTType::T_UTF8STRING:
                    (*(nsACString **)param)->Truncate();
                    (*(nsACString **)param)->SetIsVoid(PR_TRUE);
                    rv = TCL_OK;
                    break;

                default:
                    if (type->IsReference())
                    {
                        // tcl error,  can't have a null return value
                        Tcl_ResetResult(interp);
                        Tcl_AppendResult(interp, "reference argument won't accept a NULL value", NULL);
                        rv = TCL_ERROR;
                    } else
                    {
                        *(void **)param = NULL;
                        rv = TCL_OK;
                    }
            }

            return rv;
        }

        switch (type->TagPart())
        {
            case nsXPTType::T_I8:
                rv = Tcl_GetIntFromObj(interp, value, &intVal);
                if (rv == TCL_OK)
                {
                    *(void **)param = (void *)memmgr->Alloc(sizeof(PRInt8));
                    **(PRInt8 **)param = (PRInt8)intVal;
                }
                break;

            case nsXPTType::T_I16:
                rv = Tcl_GetIntFromObj(interp, value, &intVal);
                if (rv == TCL_OK)
                {
                    *(void **)param = (void *)memmgr->Alloc(sizeof(PRInt16));
                    **(PRInt16 **)param = (PRInt16)intVal;
                }
                break;

            case nsXPTType::T_I32:
                rv = Tcl_GetLongFromObj(interp, value, &longVal);
                if (rv == TCL_OK)
                {
                    *(void **)param = (void *)memmgr->Alloc(sizeof(PRInt32));
                    **(PRInt32 **)param = (PRInt32)longVal;
                }
                break;

            case nsXPTType::T_I64:
                rv = TclXPCOM_GetWideIntFromObj(interp, value, &wideVal);
                if (rv == TCL_OK)
                {
                    *(void **)param = (void *)memmgr->Alloc(sizeof(PRInt64));
                    **(PRInt64 **)param = (PRInt64)wideVal;
                }
                break;

            case nsXPTType::T_U8:
                rv = Tcl_GetIntFromObj(interp, value, &intVal);
                if (rv == TCL_OK)
                {
                    *(void **)param = (void *)memmgr->Alloc(sizeof(PRUint8));
                    **(PRUint8 **)param = (PRUint8)intVal;
                }
                break;

            case nsXPTType::T_U16:
                rv = Tcl_GetIntFromObj(interp, value, &intVal);
                if (rv == TCL_OK)
                {
                    *(void **)param = (void *)memmgr->Alloc(sizeof(PRUint16));
                    **(PRUint16 **)param = (PRUint16)intVal;
                }
                break;

            case nsXPTType::T_U32:
                rv = Tcl_GetLongFromObj(interp, value, &longVal);
                if (rv == TCL_OK)
                {
                    *(void **)param = (void *)memmgr->Alloc(sizeof(PRUint32));
                    **(PRUint32 **)param = (PRUint32)longVal;
                }
                break;

            case nsXPTType::T_U64:
                rv = TclXPCOM_GetWideIntFromObj(interp, value, &wideVal);
                if (rv == TCL_OK)
                {
                    *(void **)param = (void *)memmgr->Alloc(sizeof(PRUint64));
                    **(PRUint64 **)param = (PRUint64)wideVal;
                }
                break;

            case nsXPTType::T_FLOAT:
                rv = Tcl_GetDoubleFromObj(interp, value, &doubleVal);
                if (rv == TCL_OK)
                {
                    *(void **)param = (void *)memmgr->Alloc(sizeof(float));
                    **(float **)param = (float)doubleVal;
                }
                break;

            case nsXPTType::T_DOUBLE:
                rv = Tcl_GetDoubleFromObj(interp, value, &doubleVal);
                if (rv == TCL_OK)
                {
                    *(void **)param = (void *)memmgr->Alloc(sizeof(double));
                    **(double **)param = doubleVal;
                }
                break;

            case nsXPTType::T_BOOL:
                rv = Tcl_GetBooleanFromObj(interp, value, &intVal);
                if (rv == TCL_OK)
                {
                    *(void **)param = (void *)memmgr->Alloc(sizeof(PRBool));
                    **(PRBool **)param = (PRBool)intVal;
                }
                break;

            case nsXPTType::T_CHAR:
                if (Tcl_GetCharLength(value) != 1)
                {
                    Tcl_ResetResult(interp);
                    Tcl_AppendResult(interp, "parameter must be a single character", NULL);
                    rv = TCL_ERROR;
                } else
                {
                    utf8str = Tcl_GetStringFromObj(value, &numbytes);
                    rv = Tcl_UtfToExternal(interp, ascii_enc, utf8str, numbytes, 0, NULL, str, 2, NULL, NULL, NULL);
                    if (rv == TCL_OK)
                    {
                        *(void **)param = (void *)memmgr->Alloc(sizeof(char));
                        **(char **)param = str[0];
                    }
                }
                break;

            case nsXPTType::T_WCHAR:
                if (Tcl_GetCharLength(value) != 1)
                {
                    Tcl_ResetResult(interp);
                    Tcl_AppendResult(interp, "parameter must be a single character", NULL);
                    rv = TCL_ERROR;
                } else
                {
                    *(void **)param = (void *)memmgr->Alloc(sizeof(PRUnichar));
                    **(PRUnichar **)param = (PRUnichar)*Tcl_GetUnicode(value);
                    rv = TCL_OK;
                }
                break;

            case nsXPTType::T_VOID:
                rv = TclXPCOM_GetPointerFromObj(interp, value, (void **)param);
                // no allocation takes place
                break;

            case nsXPTType::T_IID:
                rv = TclXPCOM_GetNsIDFromObj(interp, value, &iidPtr);
                if (rv == TCL_OK)
                {
                    *(nsIID **)param = (nsIID *)CloneMemory(iidPtr, sizeof(nsIID));
                }
                break;
    
            case nsXPTType::T_ASTRING:
            case nsXPTType::T_DOMSTRING:
                (*(nsAString **)param)->Assign(Tcl_GetUnicode(value), Tcl_GetCharLength(value));
                rv = TCL_OK;
                break;

            case nsXPTType::T_CSTRING:
                length = (PRUint32)Tcl_GetCharLength(value);
                charPtr = (char *)memmgr->Alloc(sizeof(char) * (length + 1));

                utf8str = Tcl_GetStringFromObj(value, &numbytes);
                rv = Tcl_UtfToExternal(interp, ascii_enc, utf8str, numbytes, 0, NULL, charPtr, (length + 1), NULL, NULL, NULL);
                if (rv == TCL_OK)
                {
                    (*(nsACString **)param)->Assign(charPtr, length);
                }

                memmgr->Free(charPtr);
                break;

            case nsXPTType::T_UTF8STRING:
                utf8str = Tcl_GetStringFromObj(value, &numbytes);
                (*(nsACString **)param)->Assign(utf8str, numbytes);
                rv = TCL_OK;
                break;

            case nsXPTType::T_CHAR_STR:
                length = (PRUint32)Tcl_GetCharLength(value);
                charPtr = (char *)memmgr->Alloc(sizeof(char) * (length + 1));

                utf8str = Tcl_GetStringFromObj(value, &numbytes);
                rv = Tcl_UtfToExternal(interp, ascii_enc, utf8str, numbytes, 0, NULL, charPtr, (length + 1), NULL, NULL, NULL);
                if (rv == TCL_OK)
                {
                    *(char **)param = charPtr;
                } else
                {
                    memmgr->Free(charPtr);
                }
                break;

            case nsXPTType::T_WCHAR_STR:
                length = (PRUint32)Tcl_GetCharLength(value);
                *(PRUnichar **)param = (PRUnichar *)CloneMemory(Tcl_GetUnicode(value), sizeof(PRUnichar) * (length + 1));
                rv = TCL_OK;
                break;

            case nsXPTType::T_INTERFACE:
                res = interfaceinfo->GetInfoForParam(methodindex, paraminfo, &tempinfo);
                NS_ASSERTION((NS_SUCCEEDED(res)), "GetInfoForParam failed");

                rv = TclXPCOM_QueryISupportsObj(interp, value, tempinfo, &isupports);
                NS_RELEASE(tempinfo);

                if (rv == TCL_OK)
                {
                    *(nsISupports **)param = isupports;
                }
                break;

            case nsXPTType::T_INTERFACE_IS:
                interfaceinfo->GetInterfaceIsArgNumberForParam(methodindex, paraminfo, &iidindex);

                if (shared)
                {
                    iidPtr = *(nsIID **)params[iidindex];

                    // XXX will the following ever fail?
                    res = manager->GetInfoForIID(iidPtr, &tempinfo);
                    NS_ASSERTION((NS_SUCCEEDED(res)), "GetInfoForIID failed");

                    rv = TclXPCOM_QueryISupportsObj(interp, value, tempinfo, &isupports);
                    NS_RELEASE(tempinfo);

                    if (rv == TCL_OK)
                    {
                        *(nsISupports **)param = isupports;
                    }
                } else
                {
                    rv = TclXPCOM_GetISupportsFromObj(interp, value, &isupports, &tempinfo);

                    if (rv == TCL_OK)
                    {
                        tempinfo->GetInterfaceIID(&iidPtr);
                        *(nsIID **)params[iidindex] = iidPtr;
                        NS_ADDREF(isupports);
                        *(nsISupports **)param = isupports;
                    }
                }
                break;

            case nsXPTType::T_ARRAY:
                interfaceinfo->GetSizeIsArgNumberForParam(methodindex, paraminfo, 0, &sizeindex);
                interfaceinfo->GetLengthIsArgNumberForParam(methodindex, paraminfo, 0, &lengthindex);

                rv = Tcl_ListObjGetElements(interp, value, &intVal, &listPtr);
                if (rv != TCL_OK)
                {
                    return rv;
                }

                length = (PRUint32)intVal;
                if (shared)
                {
                    if (*(PRUint32 *)params[lengthindex] != length)
                    {
                        sprintf(str, "list parameter must be of length %d", length);
                        Tcl_ResetResult(interp);
                        Tcl_AppendResult(interp, str, NULL);
                        return TCL_ERROR;
                    }
                } else
                {
                    *(PRUint32 *)params[sizeindex]   = length;
                    *(PRUint32 *)params[lengthindex] = length;
                }

                if (length)
                {
                    interfaceinfo->GetTypeForParam(methodindex, paraminfo, 1, &arraytype);
                    elementsize = GetElementTypeSize(&arraytype);

                    charPtr = (char *)memmgr->Alloc(elementsize * length);
                    memset(charPtr, 0, elementsize * length); // zero it so it can be cleaned up if there's an error

                    *(void **)param = charPtr;

                    for (i = 0; i < length; i++)
                    {
                        rv = Tcl2Native(interp, listPtr[i], params, (void *)(charPtr + (i * elementsize)), &arraytype, 0, interfaceinfo, methodindex, paraminfo);
                        if (rv != TCL_OK) {
                            sprintf(str, "\nwhile reading list element #%d", i);
                            Tcl_AddObjErrorInfo(interp, str, -1);
                            // dont free charPtr, it will be cleaned up later
                            break;
                        }
                    }
                } else
                {
                    // array is empty
                    *(void **)param = NULL;
                }
                break;

            case nsXPTType::T_PSTRING_SIZE_IS:
                interfaceinfo->GetSizeIsArgNumberForParam(methodindex, paraminfo, 0, &sizeindex);
                interfaceinfo->GetLengthIsArgNumberForParam(methodindex, paraminfo, 0, &lengthindex);

                length = (PRUint32)Tcl_GetCharLength(value);
                if (shared)
                {
                    if (*(PRUint32 *)params[lengthindex] != length)
                    {
                        sprintf(str, "string parameter must be of length %d", length);
                        Tcl_ResetResult(interp);
                        Tcl_AppendResult(interp, str, NULL);
                        return TCL_ERROR;
                    }
                } else
                {
                    *(PRUint32 *)params[sizeindex]   = length;
                    *(PRUint32 *)params[lengthindex] = length;
                }
                
                charPtr = (char *)memmgr->Alloc(sizeof(char) * (length + 1));

                utf8str = Tcl_GetStringFromObj(value, &numbytes);
                rv = Tcl_UtfToExternal(interp, ascii_enc, utf8str, numbytes, 0, NULL, charPtr, (length + 1), NULL, NULL, NULL);
                if (rv == TCL_OK)
                {
                    *(char **)param = charPtr;
                } else
                {
                    memmgr->Free(charPtr);
                }
                break;

            case nsXPTType::T_PWSTRING_SIZE_IS:
                interfaceinfo->GetSizeIsArgNumberForParam(methodindex, paraminfo, 0, &sizeindex);
                interfaceinfo->GetLengthIsArgNumberForParam(methodindex, paraminfo, 0, &lengthindex);

                length = (PRUint32)Tcl_GetCharLength(value);
                if (shared)
                {
                    if (*(PRUint32 *)params[lengthindex] != length)
                    {
                        sprintf(str, "unicode string parameter must be of length %d", length);
                        Tcl_ResetResult(interp);
                        Tcl_AppendResult(interp, str, NULL);
                        return TCL_ERROR;
                    }
                } else
                {
                    *(PRUint32 *)params[sizeindex]   = length;
                    *(PRUint32 *)params[lengthindex] = length;
                }

                *(PRUnichar **)param = (PRUnichar *)CloneMemory(Tcl_GetUnicode(value), sizeof(PRUnichar) * (length + 1));
                rv = TCL_OK;
                break;

            default:
                // xpcom error,  invalid type
                TclXPCOM_SetError(interp, NS_ERROR_INVALID_ARG);
                rv = TCL_ERROR;
                break;
        }
    }
    return rv;
}

