/* -*- Mode: C++; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is TclXPCOM.
 * 
 * The Initial Developer of the Original Code is Mark Follett.
 * Portions created by Mark Follett are Copyright (C) 2001-2002
 * Mark Follett.  All Rights Reserved.
 * 
 * Contributor(s):
 *     Mark Follett <mef123@myrealbox.com> (Original Author)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 */

/**
 *  @file TclXPCOMObjType.cpp
 *  Contains functions for handling custom TclXPCOM objects.
 */

#include "TclXPCOMPrivate.h"

#ifndef USE_TCL_WIDEINT
#   include "prprf.h"
#endif

#define NULL_STRINGREP "TclXPCOM NULL Object"

/**
 *  The following are structures used by the Tcl 'cmdName'
 *  object type, which caches command lookups. We are hijacking
 *  this type in order to allow the TclXPCOM isupports object
 *  to act as a command. This provides for oo-style method
 *  invocation, like this:
 *
 *    [$component QueryInterface nsISupports]
 *
 *  instead of this:
 *
 *    [xpcom::invoke $component QueryInterface nsISupports]
 *
 *  (both of the above are equivalent)
 *
 *      typedef struct Command
 *      {
 *          Tcl_HashEntry *hPtr;
 *          Namespace *nsPtr;
 *          int refCount;
 *          int cmdEpoch;
 *          CompileProc *compileProc;
 *          Tcl_ObjCmdProc *objProc;
 *          ClientData objClientData;
 *          Tcl_CmdProc *proc;
 *          ClientData clientData;
 *          Tcl_CmdDeleteProc *deleteProc;
 *          ClientData deleteData;
 *          int flags;
 *          ImportRef *importRefPtr;
 *          CommandTrace *tracePtr;
 *      } Command;
 *
 *      typedef struct ResolvedCmdName
 *      {
 *          Command *cmdPtr;
 *          Namespace *refNsPtr;
 *          long refNsId;
 *          int refNsCmdEpoch;
 *          int cmdEpoch;
 *          int refCount;
 *      } ResolvedCmdName;
 */

typedef struct TclXPCOM_Command
{
    char           *dummy1;
    char           *dummy2;
    int             dummy3;
    int             dummy4;
    char           *dummy5;
    Tcl_ObjCmdProc *objProc;
    ClientData      objClientData;
    char           *dummy6;
    ClientData      dummy7;
    char           *dummy8;
    ClientData      dummy9;
    int             dummy10;
    char           *dummy11;
    char           *dummy12;
} TclXPCOM_Command;

typedef struct TclXPCOM_ResolvedCmdName
{
    TclXPCOM_Command *cmdPtr;
    char *dummy1;
    long dummy2;
    int dummy3;
    int dummy4;
    int dummy5;
} TclXPCOM_ResolvedCmdName;

static TclXPCOM_Command isupportscmd =
{
    NULL,
    NULL,
    0,
    0,
    NULL,
    TclXPCOM_InvokeObjCmd,
    (ClientData)0,
    NULL,
    NULL,
    NULL,
    NULL,
    0,
    NULL,
    NULL
};

static TclXPCOM_ResolvedCmdName resolvedisupports =
{
    &isupportscmd,
    NULL,
    0,
    0,
    0,
    0
};

Tcl_ObjType ISupportsType;

static Tcl_ObjType SavedCmdNameType;
static Tcl_ObjType NsIDType;
static Tcl_ObjType PointerType;
static Tcl_ObjType IdentifierType;
static Tcl_ObjType *CmdNameType;

#ifndef USE_TCL_WIDEINT
#define TCLXPCOM_WIDEINTTYPENAME "TclXPCOM_WideInt"
static Tcl_ObjType WideIntType;
#endif

/**
 *  Sets the string representation of a TclXPCOM pointer object.
 */
static void
UpdateStringOfPointer(
    Tcl_Obj *objPtr ///< (in) Tcl object to set string rep
)
{
    int length;
    char str[128];

    NS_PRECONDITION(objPtr, "null pointer");

    if (objPtr->internalRep.twoPtrValue.ptr1 == NULL)
    {
        length = sprintf(str, NULL_STRINGREP);
    } else
    {
        length = sprintf(str, "TclXPCOM Pointer @%p", objPtr->internalRep.twoPtrValue.ptr1);
    }

    objPtr->bytes = Tcl_Alloc(length + 1);
    memcpy(objPtr->bytes, str, length + 1);
    objPtr->length = length;
}


/**
 *  Attempts to convert a Tcl object to a TclXPCOM pointer
 *  object based on its string representation. If 'interp' is
 *  not NULL, then its result will be set to any error messages
 *  if the conversion failed.
 *  
 *  @retval TCL_OK    - if conversion was successful
 *  @retval TCL_ERROR - if an error occurred
 */
static int
SetPointerFromAny(
    Tcl_Interp *interp,///< (in) interpreter to store any error messages
    Tcl_Obj    *objPtr ///< (in) Tcl object to convert
)
{
    int rv;
    void *ptr;

    // no precondition on interp
    NS_PRECONDITION(objPtr, "null pointer");

    if (TclXPCOM_IsNULL(objPtr))
    {
        ptr = NULL;
        rv = TCL_OK;

    } else if (sscanf(Tcl_GetString(objPtr), "TclXPCOM Pointer @%p", (void **)&ptr) == 1)
    {
        rv = TCL_OK;
    } else
    {
        if (interp)
        {
            Tcl_ResetResult(interp);
            Tcl_AppendResult(interp, "invalid pointer object \"", Tcl_GetString(objPtr), "\"", NULL);
        }
        rv = TCL_ERROR;
    }

    if (rv == TCL_OK)
    {
        if (objPtr->typePtr && objPtr->typePtr->freeIntRepProc)
        {
            objPtr->typePtr->freeIntRepProc(objPtr);
        }
        objPtr->internalRep.twoPtrValue.ptr1 = ptr;
        objPtr->typePtr = &PointerType;
    }

    return rv;
}


/**
 *  Frees a TclXPCOM interface reference object.
 */
static void
FreeISupportsObj(
    Tcl_Obj *objPtr ///< (in) Tcl object to free
)
{
    Tcl_HashEntry *entry;

    NS_PRECONDITION(objPtr, "null pointer");

//  printf("Freeing: %p\n", objPtr);
    entry = (Tcl_HashEntry *)objPtr->internalRep.twoPtrValue.ptr2;
    ReleaseInterface((nsISupports *)Tcl_GetHashValue(entry));
}


/**
 *  Duplicates a TclXPCOM interface reference object.
 */
static void
DuplicateISupportsObj(
    Tcl_Obj *srcPtr,///< (in) source Tcl object
    Tcl_Obj *dupPtr ///< (in) duplicate Tcl object
)
{
    Tcl_HashEntry *entry;

    NS_PRECONDITION(srcPtr, "null pointer");
    NS_PRECONDITION(dupPtr, "null pointer");

//  printf("Duplicating from: %p to %p\n", srcPtr, dupPtr);

    dupPtr->internalRep.twoPtrValue = srcPtr->internalRep.twoPtrValue;
    entry = (Tcl_HashEntry *)dupPtr->internalRep.twoPtrValue.ptr2;
    AddRefInterface((nsISupports *)Tcl_GetHashValue(entry), (nsIInterfaceInfo *)entry->key.oneWordValue, NULL);
}


/**
 *  Sets the string representation of a TclXPCOM interface
 *  reference object.
 */
static void
UpdateStringOfISupports(
    Tcl_Obj *objPtr ///< (in) Tcl object to set string rep
)
{
    int length;
    char str[128];
    const char *name;
    Tcl_HashEntry *entry;

    NS_PRECONDITION(objPtr, "null pointer");

//  printf("Updating: %p\n", objPtr);

    entry = (Tcl_HashEntry *)objPtr->internalRep.twoPtrValue.ptr2;
    ((nsIInterfaceInfo *)(entry->key.oneWordValue))->GetNameShared(&name);

    length = sprintf(str, "xpcom::inscope %s::%p", name, Tcl_GetHashValue(entry));

    objPtr->bytes = Tcl_Alloc(length + 1);
    memcpy(objPtr->bytes, str, length + 1);
    objPtr->length = length;
}


/**
 *  Attempts to convert a Tcl object to a TclXPCOM interface
 *  reference object based on its string representation. If
 *  'interp' is not NULL, then its result will be set to any
 *  error messages if the conversion failed.
 *  
 *  @retval TCL_OK    - if conversion was successful
 *  @retval TCL_ERROR - if an error occurred
 */
static int
SetISupportsFromAny(
    Tcl_Interp *interp,///< (in) interpreter to store any error messages
    Tcl_Obj    *objPtr ///< (in) Tcl object to convert
)
{
    int rv;
    nsresult res;
    char name[128];
    nsISupports *ref, *isupports;
    nsIInterfaceInfo *interfaceinfo;
    InterfaceRef *iref;
    Tcl_HashEntry *entry;

    // no precondition on interp
    NS_PRECONDITION(objPtr, "null pointer");

    if (objPtr->typePtr == CmdNameType && objPtr->internalRep.twoPtrValue.ptr2)
    {
        return TCL_OK;
    }

//  printf("Setting from: %p\n", objPtr);

    if (sscanf(Tcl_GetString(objPtr), "xpcom::inscope %127[^:]::%p", name, (void **)&ref) == 2)
    {
        iref = GetInterfaceRef(ref);
        if (iref)
        {
            res = manager->GetInfoForName(name, &interfaceinfo);
            if (NS_SUCCEEDED(res))
            {
                rv = QueryInterfaceIRef(NULL, iref, interfaceinfo, &isupports);
                if (rv != TCL_OK)
                {
                    if (interp)
                    {
                        Tcl_ResetResult(interp);
                        Tcl_AppendResult(interp, "interface \"", name, "\" is not supported by \"", Tcl_GetString(objPtr), "\"", NULL);
                    }
                }
            } else
            {
                if (interp)
                {
                    Tcl_ResetResult(interp);
                    Tcl_AppendResult(interp, "unknown interface \"", name, "\" in interface reference \"", Tcl_GetString(objPtr), "\"", NULL);
                }
                rv = TCL_ERROR;
            }
        } else
        {
            if (interp)
            {
                Tcl_ResetResult(interp);
                Tcl_AppendResult(interp, "interface reference \"", Tcl_GetString(objPtr), "\" doesn't point to a known component", NULL);
            }
            rv = TCL_ERROR;
        }
    } else
    {
        if (interp)
        {
            Tcl_ResetResult(interp);
            Tcl_AppendResult(interp, "invalid interface reference \"", Tcl_GetString(objPtr), "\"", NULL);
        }
        rv = TCL_ERROR;
    }

    if (rv == TCL_OK)
    {
        if (objPtr->typePtr && objPtr->typePtr->freeIntRepProc)
        {
            objPtr->typePtr->freeIntRepProc(objPtr);
        }
        objPtr->internalRep.twoPtrValue.ptr1 = &resolvedisupports;
        AddRefInterface(isupports, interfaceinfo, &entry);
        objPtr->internalRep.twoPtrValue.ptr2 = entry;
        objPtr->typePtr = CmdNameType;
    }

    return rv;
}


/**
 *  Replacement for the built-in FreeCmdNameInternalRep function.
 */
static void
FreeCmdNameObj(
    Tcl_Obj *objPtr ///< (in) Tcl object to free
)
{
    NS_PRECONDITION(objPtr, "null pointer");

    if (objPtr->internalRep.twoPtrValue.ptr2)
    {
        FreeISupportsObj(objPtr);
    } else
    {
        SavedCmdNameType.freeIntRepProc(objPtr);
    }
}


/**
 *  Replacement for the built-in DupCmdNameInternalRep function.
 */
static void
DuplicateCmdNameObj(
    Tcl_Obj *srcPtr,///< (in) source Tcl object
    Tcl_Obj *dupPtr ///< (in) duplicate Tcl object
)
{
    NS_PRECONDITION(srcPtr, "null pointer");
    NS_PRECONDITION(dupPtr, "null pointer");

    if (srcPtr->internalRep.twoPtrValue.ptr2)
    {
        DuplicateISupportsObj(srcPtr, dupPtr);
    } else
    {
        SavedCmdNameType.dupIntRepProc(srcPtr, dupPtr);
    }
}


/**
 *  Replacement for the (non-existent) built-in UpdateStringOfCmdName function.
 */
static void
UpdateStringOfCmdName(
    Tcl_Obj *objPtr ///< (in) Tcl object to set string rep
)
{
    NS_PRECONDITION(objPtr, "null pointer");

    if (objPtr->internalRep.twoPtrValue.ptr2)
    {
        UpdateStringOfISupports(objPtr);
    } else
    {
        NS_ASSERTION(0, "shouldn't call UpdateString on a cmdName object");
    }
}


/**
 *  Replacement for the built-in SetCmdNameFromAny function.
 *  
 *  @retval TCL_OK    - if conversion was successful
 *  @retval TCL_ERROR - if an error occurred
 */
static int
SetCmdNameFromAny(
    Tcl_Interp *interp,///< (in) interpreter to store any error messages
    Tcl_Obj    *objPtr ///< (in) Tcl object to convert
)
{
    int rv;

    // no precondition on interp
    NS_PRECONDITION(objPtr, "null pointer");

    rv = SetISupportsFromAny(NULL, objPtr);

    if (rv != TCL_OK)
    {
        rv = SavedCmdNameType.setFromAnyProc(interp, objPtr);
    }

    return rv;
}


/**
 *  Frees a TclXPCOM identifier object.
 */
static void
FreeIdentifierObj(
    Tcl_Obj *objPtr ///< (in) Tcl object to free
)
{
    NS_PRECONDITION(objPtr, "null pointer");

    nsIInterfaceInfo *interfaceinfo;

    if (objPtr->internalRep.twoPtrValue.ptr1)
    {
        interfaceinfo = (nsIInterfaceInfo *)objPtr->internalRep.twoPtrValue.ptr1;
        NS_RELEASE(interfaceinfo);
    }

    memmgr->Free(objPtr->internalRep.twoPtrValue.ptr2);
}


/**
 *  Duplicates a TclXPCOM identifier object.
 */
static void
DuplicateIdentifierObj(
    Tcl_Obj *srcPtr,///< (in) source Tcl object
    Tcl_Obj *dupPtr ///< (in) duplicate Tcl object
)
{
    NS_PRECONDITION(srcPtr, "null pointer");
    NS_PRECONDITION(dupPtr, "null pointer");

    if (srcPtr->internalRep.twoPtrValue.ptr1)
    {
        dupPtr->internalRep.twoPtrValue.ptr1 = srcPtr->internalRep.twoPtrValue.ptr1;
        NS_ADDREF(((nsIInterfaceInfo *)dupPtr->internalRep.twoPtrValue.ptr1));
    } else
    {
        dupPtr->internalRep.twoPtrValue.ptr1 = NULL;
    }

    dupPtr->internalRep.twoPtrValue.ptr2 = (void *)CloneMemory(srcPtr->internalRep.twoPtrValue.ptr2, sizeof(char) * (strlen((char *)srcPtr->internalRep.twoPtrValue.ptr2) + 1));
}


/**
 *  Sets the string representation of a TclXPCOM identifier object.
 */
static void
UpdateStringOfIdentifier(
    Tcl_Obj *objPtr ///< (in) Tcl object to set string rep
)
{
    int inamelength, identifierlength, length;
    const char *iname;

    NS_PRECONDITION(objPtr, "null pointer");

    if (objPtr->internalRep.twoPtrValue.ptr1)
    {
        ((nsIInterfaceInfo *)objPtr->internalRep.twoPtrValue.ptr1)->GetNameShared(&iname);

        inamelength = strlen(iname);
        identifierlength = strlen((char *)objPtr->internalRep.twoPtrValue.ptr2);

        length = inamelength + 2 + identifierlength;

        objPtr->bytes = Tcl_Alloc(length + 1);
        memcpy(objPtr->bytes, iname, inamelength);
        objPtr->bytes[inamelength] = ':';
        objPtr->bytes[inamelength + 1] = ':';
        memcpy(&objPtr->bytes[inamelength + 2], objPtr->internalRep.twoPtrValue.ptr2, identifierlength + 1);
        objPtr->length = length;

    } else
    {
        length = strlen((char *)objPtr->internalRep.twoPtrValue.ptr2);

        objPtr->bytes = Tcl_Alloc(length + 1);
        memcpy(objPtr->bytes, objPtr->internalRep.twoPtrValue.ptr2, length + 1);
        objPtr->length = length;
    }
}


/**
 *  Attempts to convert a Tcl object to a TclXPCOM identifier
 *  object based on its string representation. The identifier
 *  may be prefixed with a known interface name. If 'interp' is
 *  not NULL, then its result will be set to any error messages
 *  if the conversion failed.
 *  
 *  @retval TCL_OK    - if conversion was successful
 *  @retval TCL_ERROR - if an error occurred
 */
static int
SetIdentifierFromAny(
    Tcl_Interp *interp,///< (in) interpreter to store any error messages
    Tcl_Obj    *objPtr ///< (in) Tcl object to convert
)
{
    int rv, index;
    nsresult res;
    char *interfacename, *identifier;
    nsIInterfaceInfo *interfaceinfo;

    // no precondition on interp
    NS_PRECONDITION(objPtr, "null pointer");

    identifier = Tcl_GetString(objPtr);

    for (index = strlen(identifier); index > 1; index--)
    {
        if (identifier[index-1] == ':' &&
            identifier[index-2] == ':')
        {
            break;
        }
    }

    if (index > 1)
    {
        interfacename = (char *)memmgr->Alloc(sizeof(char) * (index - 1));
        memcpy(interfacename, identifier, sizeof(char) * (index - 2));
        interfacename[index - 2] = '\0';

        res = manager->GetInfoForName(interfacename, &interfaceinfo);

        if (NS_FAILED(res))
        {
            if (interp)
            {
                Tcl_ResetResult(interp);
                Tcl_AppendResult(interp, "unknown interface \"", interfacename, "\" in identifier \"", identifier, "\"", NULL);
            }
            rv = TCL_ERROR;
        } else
        {
            rv = TCL_OK;
        }

        memmgr->Free(interfacename);

    } else
    {
        index = 0;
        interfaceinfo = NULL;
        rv = TCL_OK;
    }

    if (rv == TCL_OK)
    {
        if (objPtr->typePtr && objPtr->typePtr->freeIntRepProc)
        {
            objPtr->typePtr->freeIntRepProc(objPtr);
        }
        objPtr->internalRep.twoPtrValue.ptr1 = interfaceinfo;
        objPtr->internalRep.twoPtrValue.ptr2 = (void *)CloneMemory(&identifier[index], sizeof(char) * (strlen(&identifier[index]) + 1));
        objPtr->typePtr = &IdentifierType;
    }

    return rv;
}


/**
 *  Frees a TclXPCOM nsID object.
 */
static void
FreeNsIDObj(
    Tcl_Obj *objPtr ///< (in) Tcl object to free
)
{
    NS_PRECONDITION(objPtr, "null pointer");

    memmgr->Free(objPtr->internalRep.otherValuePtr);
}


/**
 *  Duplicates a TclXPCOM nsID object.
 */
static void
DuplicateNsIDObj(
    Tcl_Obj *srcPtr,///< (in) source Tcl object
    Tcl_Obj *dupPtr ///< (in) duplicate Tcl object
)
{
    NS_PRECONDITION(srcPtr, "null pointer");
    NS_PRECONDITION(dupPtr, "null pointer");

    dupPtr->internalRep.otherValuePtr = (void *)CloneMemory(srcPtr->internalRep.otherValuePtr, sizeof(nsID));
}


/**
 *  Sets the string representation of a TclXPCOM nsID object.
 */
static void
UpdateStringOfNsID(
    Tcl_Obj *objPtr ///< (in) Tcl object to set string rep
)
{
    int length;
    char *str;
    nsresult res;

    NS_PRECONDITION(objPtr, "null pointer");

    if ((res = manager->GetNameForIID((nsIID *)objPtr->internalRep.otherValuePtr, &str)) == NS_OK)
    {
        // type is IID
    }
    else if ((res = registrar->CIDToContractID(*(nsCID *)objPtr->internalRep.otherValuePtr, &str)) == NS_OK)
    {
        // type is CID
    }
    else
    {
        // type is ID
        str = ((nsID *)objPtr->internalRep.otherValuePtr)->ToString();
    }

    length = strlen(str);

    objPtr->bytes = Tcl_Alloc(length + 1);
    memcpy(objPtr->bytes, str, length + 1);
    objPtr->length = length;

    memmgr->Free(str);
}


/**
 *  Attempts to convert a Tcl object to a TclXPCOM nsID object
 *  based on its string representation. The object's string
 *  representation may match any known interface name, contract
 *  id, or raw unique id string. If 'interp' is not NULL, then
 *  its result will be set to any error messages if the conversion
 *  failed.
 *  
 *  @retval TCL_OK    - if conversion was successful
 *  @retval TCL_ERROR - if an error occurred
 */
static int
SetNsIDFromAny(
    Tcl_Interp *interp,///< (in) interpreter to store any error messages
    Tcl_Obj    *objPtr ///< (in) Tcl object to convert
)
{
    nsID id_stack;
    nsID *id;
    nsresult res;

    // no precondition on interp
    NS_PRECONDITION(objPtr, "null pointer");

    if (id_stack.Parse(Tcl_GetString(objPtr)))
    {
        // type is ID
        id = (nsID *)CloneMemory(&id_stack, sizeof(nsID));
    } else if ((res = manager->GetIIDForName(Tcl_GetString(objPtr), (nsIID **)&id)) == NS_OK)
    {
        // type is IID
    } else if ((res = registrar->ContractIDToCID(Tcl_GetString(objPtr), (nsCID **)&id)) == NS_OK)
    {
        // type is CID
    } else
    {
        if (interp)
        {
            Tcl_ResetResult(interp);
            Tcl_AppendResult(interp, "invalid ID string \"", Tcl_GetString(objPtr), "\"", NULL);
        }
        return TCL_ERROR;
    }

    if (objPtr->typePtr && objPtr->typePtr->freeIntRepProc)
    {
        objPtr->typePtr->freeIntRepProc(objPtr);
    }
    objPtr->internalRep.otherValuePtr = id;
    objPtr->typePtr = &NsIDType;
    
    return TCL_OK;
}


#ifndef USE_TCL_WIDEINT

/**
 *  Sets the string representation of a TclXPCOM 64-bit integer object.
 */
static void
UpdateStringOfWideInt(
    Tcl_Obj *objPtr ///< (in) Tcl object to set string rep
)
{
    int length;
    char str[128];

    NS_PRECONDITION(objPtr, "null pointer");

    length = PR_snprintf(str, 128, "%lld", *(PRInt64 *)(&objPtr->internalRep));

    objPtr->bytes = Tcl_Alloc(length + 1);
    memcpy(objPtr->bytes, str, length + 1);
    objPtr->length = length;
}


/**
 *  Attempts to convert a Tcl object to a TclXPCOM 64-bit integer
 *  object based on its string representation. If 'interp' is not
 *  NULL, then its result will be set to any error messages if
 *  the conversion failed.
 *  
 *  @retval TCL_OK    - if conversion was successful
 *  @retval TCL_ERROR - if an error occurred
 */
static int
SetWideIntFromAny(
    Tcl_Interp *interp,///< (in) interpreter to store any error messages
    Tcl_Obj    *objPtr ///< (in) Tcl object to convert
)
{
    int rv;
    PRInt64 wideint;

    // no precondition on interp
    NS_PRECONDITION(objPtr, "null pointer");

    if (PR_sscanf(Tcl_GetString(objPtr), "%lld", &wideint))
    {
        rv = TCL_OK;
    } else
    {
        if (interp)
        {
            Tcl_ResetResult(interp);
            Tcl_AppendResult(interp, "invalid 64-bit integer object \"", Tcl_GetString(objPtr), "\"", NULL);
        }
        rv = TCL_ERROR;
    }

    if (rv == TCL_OK)
    {
        if (objPtr->typePtr && objPtr->typePtr->freeIntRepProc)
        {
            objPtr->typePtr->freeIntRepProc(objPtr);
        }
        *(PRInt64 *)(&objPtr->internalRep) = wideint;
        objPtr->typePtr = &WideIntType;
    }

    return rv;
}


/**
 *  Creates a new TclXPCOM 64-bit integer object.
 *  
 *  @return a new 64-bit integer object with refcount 0
 */
Tcl_Obj *
TclXPCOM_NewWideIntObj(
    PRInt64 wideint ///< (in) 64-bit integer
)
{
    Tcl_Obj *result;

    result = Tcl_NewObj();

    result->typePtr = &WideIntType;
    *(PRInt64 *)(&result->internalRep) = wideint;
    Tcl_InvalidateStringRep(result);

    return result;
}


/**
 *  Retrieves the TclXPCOM 64-bit integer internal
 *  representation from a Tcl object. If an error occurs,
 *  a message will be left in 'interp's result unless 'interp'
 *  is NULL.
 *  
 *  @retval TCL_OK    - if successful
 *  @retval TCL_ERROR - if an error occurred
 */
int
TclXPCOM_GetWideIntFromObj(
    Tcl_Interp *interp, ///< (in)  interpreter to store any error messages
    Tcl_Obj    *objPtr, ///< (in)  Tcl object
    PRInt64    *wideint ///< (out) address to store the 64-bit integer
)
{
    int rv;

    // no precondition on interp
    NS_PRECONDITION(objPtr, "null pointer");
    NS_PRECONDITION(wideint, "null pointer");

    rv = Tcl_ConvertToType(interp, objPtr, &WideIntType);

    if (rv == TCL_OK)
    {
        *wideint = *(PRInt64 *)(&objPtr->internalRep);
    }

    return rv;
}

#endif /* USE_TCL_WIDEINT */


/**
 *  Creates a new TclXPCOM interface reference object that
 *  holds a new Addrefed interface pointer.
 *  
 *  @return a new interface reference object with refcount 0
 */
Tcl_Obj *
TclXPCOM_NewISupportsObj(
    nsISupports      *isupports,    ///< (in) interface pointer
    nsIInterfaceInfo *interfaceinfo ///< (in) interface info
)
{
    Tcl_Obj *result;
    Tcl_HashEntry *entry;

    NS_PRECONDITION(isupports, "null pointer");
    NS_PRECONDITION(interfaceinfo, "null pointer");

    result = Tcl_NewObj();

    result->typePtr = CmdNameType;
    result->internalRep.twoPtrValue.ptr1 = &resolvedisupports;
    AddRefInterface(isupports, interfaceinfo, &entry);
    result->internalRep.twoPtrValue.ptr2 = entry;
    Tcl_InvalidateStringRep(result);

    return result;
}


/**
 *  Creates a new TclXPCOM nsID object that holds a copy
 *  of the given nsID.
 *  
 *  @return a new nsID object with refcount 0
 */
Tcl_Obj *
TclXPCOM_NewNsIDObj(
    nsID *id ///< (in) nsID to be held by the object
)
{
    Tcl_Obj *result;

    NS_PRECONDITION(id, "null pointer");

    result = Tcl_NewObj();

    result->typePtr = &NsIDType;
    result->internalRep.otherValuePtr = (void *)CloneMemory(id, sizeof(nsID));
    Tcl_InvalidateStringRep(result);

    return result;
}


/**
 *  Creates a new TclXPCOM identifier object that holds a copy
 *  of the given interface info and identifier.
 *  
 *  @return a new identifier object with refcount 0
 */
Tcl_Obj *
TclXPCOM_NewIdentifierObj(
    nsIInterfaceInfo *interfaceinfo,///< (in) interface info of identifier
    char             *identifier    ///< (in) name of identifier
)
{
    Tcl_Obj *result;

    NS_PRECONDITION(interfaceinfo, "null pointer");
    NS_PRECONDITION(identifier, "null pointer");

    result = Tcl_NewObj();

    result->typePtr = &IdentifierType;
    result->internalRep.twoPtrValue.ptr1 = interfaceinfo;
    NS_ADDREF(interfaceinfo);
    result->internalRep.twoPtrValue.ptr2 = (char *)CloneMemory(identifier, sizeof(char) * (strlen(identifier) + 1));
    Tcl_InvalidateStringRep(result);

    return result;
}


/**
 *  Creates a new TclXPCOM pointer object.
 *  
 *  @return a new pointer object with refcount 0
 */
Tcl_Obj *
TclXPCOM_NewPointerObj(
    void *ptr ///< (in) pointer value to be held by the object
)
{
    Tcl_Obj *result;

    // no precondition on ptr

    result = Tcl_NewObj();

    result->typePtr = &PointerType;
    result->internalRep.twoPtrValue.ptr1 = ptr;
    Tcl_InvalidateStringRep(result);

    return result;
}


/**
 *  Retrieves the TclXPCOM pointer internal representation
 *  from a Tcl object. If an error occurs, a message will
 *  be left in 'interp's result unless 'interp' is NULL.
 *  
 *  @retval TCL_OK    - if successful
 *  @retval TCL_ERROR - if an error occurred
 */
int
TclXPCOM_GetPointerFromObj(
    Tcl_Interp *interp, ///< (in)  interpreter to store any error messages
    Tcl_Obj    *objPtr, ///< (in)  Tcl object
    void      **ptr     ///< (out) address to store the pointer
)
{
    int rv;

    // no precondition on interp
    NS_PRECONDITION(objPtr, "null pointer");
    NS_PRECONDITION(ptr, "null pointer");

    rv = Tcl_ConvertToType(interp, objPtr, &PointerType);

    if (rv == TCL_OK)
    {
        *ptr = objPtr->internalRep.twoPtrValue.ptr1;
    }

    return rv;
}


/**
 *  Predicate that determines whether a Tcl object is 'NULL'
 *  
 *  @retval 1 - the Tcl object is NULL
 *  @retval 0 - the Tcl object is not NULL
 */
int
TclXPCOM_IsNULL(
    Tcl_Obj *objPtr ///< (in) Tcl object
)
{
    NS_PRECONDITION(objPtr, "null pointer");

    if ((objPtr->typePtr == &PointerType) &&
        (objPtr->internalRep.twoPtrValue.ptr1) == NULL)
    {
        return 1;
    }

    if (!strcmp(Tcl_GetString(objPtr), NULL_STRINGREP))
    {
        return 1;
    }

    return 0;
}


/**
 *  Retrieves the TclXPCOM interface reference internal
 *  representation from a Tcl object. If an error occurs, a
 *  message will be left in 'interp's result unless 'interp'
 *  is NULL.
 *  
 *  @retval TCL_OK    - if successful
 *  @retval TCL_ERROR - if an error occurred
 */
int
TclXPCOM_GetISupportsFromObj(
    Tcl_Interp        *interp,       ///< (in)  interpreter to store any error messages
    Tcl_Obj           *objPtr,       ///< (in)  Tcl object
    nsISupports      **isupports,    ///< (out) address to store interface pointer
    nsIInterfaceInfo **interfaceinfo ///< (out) address to store the interface info
)
{
    int rv;
    Tcl_HashEntry *entry;

    // no precondition on interp
    NS_PRECONDITION(objPtr, "null pointer");
    NS_PRECONDITION(isupports, "null pointer");
    NS_PRECONDITION(interfaceinfo, "null pointer");

    rv = Tcl_ConvertToType(interp, objPtr, &ISupportsType);

    if (rv == TCL_OK)
    {
        entry = (Tcl_HashEntry *)objPtr->internalRep.twoPtrValue.ptr2;

        *isupports     = (nsISupports      *)Tcl_GetHashValue(entry);
        *interfaceinfo = (nsIInterfaceInfo *)entry->key.oneWordValue;
    }

    return rv;
}


/**
 *  Attempts to retrieve an interface pointer from the given
 *  Tcl object that matches the given interface. 'result'
 *  will be set to an AddRefed interface pointer which must
 *  eventually be Released. If an error occurs, a message will
 *  be left in 'interp's result unless 'interp' is NULL.
 *  
 *  @retval TCL_OK    - if successful
 *  @retval TCL_ERROR - if an error occurred
 */
int
TclXPCOM_QueryISupportsObj(
    Tcl_Interp       *interp,   ///< (in)  interpreter to store any error messages
    Tcl_Obj          *objPtr,   ///< (in)  Tcl object
    nsIInterfaceInfo *queryinfo,///< (in)  requested interface
    nsISupports     **result    ///< (out) address to store AddRefed interface pointer
)
{
    int rv;
    nsISupports *isupports;
    nsIInterfaceInfo *interfaceinfo;
    InterfaceRef *iref;

    // no precondition on interp
    NS_PRECONDITION(objPtr, "null pointer");
    NS_PRECONDITION(queryinfo, "null pointer");
    NS_PRECONDITION(result, "null pointer");

    rv = TclXPCOM_GetISupportsFromObj(interp, objPtr, &isupports, &interfaceinfo);
    if (rv != TCL_OK)
    {
        return rv;
    }

    iref = GetInterfaceRef(isupports);
    NS_ASSERTION(iref, "isupports entry doesn't exist in irefhash");

    rv = QueryInterfaceIRef(interp, iref, queryinfo, result);
    if (rv == TCL_OK)
    {
        NS_ADDREF(*result);
    }

    return rv;
}


/**
 *  Retrieves the TclXPCOM identifier internal representation
 *  from a Tcl object. If an error occurs, a message will be
 *  left in 'interp's result unless 'interp' is NULL.
 *  
 *  @retval TCL_OK    - if successful
 *  @retval TCL_ERROR - if an error occurred
 */
int
TclXPCOM_GetIdentifierFromObj(
    Tcl_Interp        *interp,       ///< (in)  interpreter to store any error messages
    Tcl_Obj           *objPtr,       ///< (in)  Tcl object
    nsIInterfaceInfo **interfaceinfo,///< (out) address to store the interface info
    char             **identifier    ///< (out) address to store the identifier name
)
{
    int rv;

    // no precondition on interp
    NS_PRECONDITION(objPtr, "null pointer");
    NS_PRECONDITION(interfaceinfo, "null pointer");
    NS_PRECONDITION(identifier, "null pointer");

    rv = Tcl_ConvertToType(interp, objPtr, &IdentifierType);

    if (rv == TCL_OK)
    {
        *interfaceinfo = (nsIInterfaceInfo *)objPtr->internalRep.twoPtrValue.ptr1;
        *identifier    = (char *)objPtr->internalRep.twoPtrValue.ptr2;
    }

    return rv;
}


/**
 *  Retrieves the nsID internal representation from a Tcl object.
 *  If an error occurs, a message will be left in 'interp's result
 *  unless 'interp' is NULL.
 *  
 *  @retval TCL_OK    - if successful
 *  @retval TCL_ERROR - if an error occurred
 */
int
TclXPCOM_GetNsIDFromObj(
    Tcl_Interp *interp,///< (in)  interpreter to store any error messages
    Tcl_Obj    *objPtr,///< (in)  Tcl object
    nsID      **id     ///< (out) address to store a pointer to the nsID
)
{
    int rv;

    // no precondition on interp
    NS_PRECONDITION(objPtr, "null pointer");
    NS_PRECONDITION(id, "null pointer");

    rv = Tcl_ConvertToType(interp, objPtr, &NsIDType);

    if (rv == TCL_OK)
    {
        *id = (nsID *)objPtr->internalRep.otherValuePtr;
    }

    return rv;
}


/**
 *  Registers the custom TclXPCOM Object types with Tcl.
 */
void
TclXPCOM_RegisterObjTypes(void)
{
    CmdNameType = Tcl_GetObjType("cmdName");
    SavedCmdNameType = *CmdNameType;

    NS_ASSERTION(CmdNameType->freeIntRepProc, "no built-in UpdateStringOfCmdName exists");
    NS_ASSERTION(CmdNameType->dupIntRepProc, "no built-in UpdateStringOfCmdName exists");
    NS_ASSERTION(!CmdNameType->updateStringProc, "built-in UpdateStringOfCmdName exists");
    NS_ASSERTION(CmdNameType->setFromAnyProc, "no built-in UpdateStringOfCmdName exists");

    CmdNameType->freeIntRepProc   = FreeCmdNameObj;
    CmdNameType->dupIntRepProc    = DuplicateCmdNameObj;
    CmdNameType->updateStringProc = UpdateStringOfCmdName;
    CmdNameType->setFromAnyProc   = SetCmdNameFromAny;

    ISupportsType.name = TCLXPCOM_ISUPPORTSTYPENAME;
    ISupportsType.freeIntRepProc   = FreeISupportsObj;
    ISupportsType.dupIntRepProc    = DuplicateISupportsObj;
    ISupportsType.updateStringProc = UpdateStringOfISupports;
    ISupportsType.setFromAnyProc   = SetISupportsFromAny;

    NsIDType.name = TCLXPCOM_NSIDTYPENAME;
    NsIDType.freeIntRepProc   = FreeNsIDObj;
    NsIDType.dupIntRepProc    = DuplicateNsIDObj;
    NsIDType.updateStringProc = UpdateStringOfNsID;
    NsIDType.setFromAnyProc   = SetNsIDFromAny;

    PointerType.name = TCLXPCOM_POINTERTYPENAME;
    PointerType.freeIntRepProc   = (Tcl_FreeInternalRepProc *)NULL;
    PointerType.dupIntRepProc    = (Tcl_DupInternalRepProc *)NULL;
    PointerType.updateStringProc = UpdateStringOfPointer;
    PointerType.setFromAnyProc   = SetPointerFromAny;

    IdentifierType.name = TCLXPCOM_IDENTIFIERTYPENAME;
    IdentifierType.freeIntRepProc   = FreeIdentifierObj;
    IdentifierType.dupIntRepProc    = DuplicateIdentifierObj;
    IdentifierType.updateStringProc = UpdateStringOfIdentifier;
    IdentifierType.setFromAnyProc   = SetIdentifierFromAny;

#ifndef USE_TCL_WIDEINT
    WideIntType.name = TCLXPCOM_WIDEINTTYPENAME;
    WideIntType.freeIntRepProc   = (Tcl_FreeInternalRepProc *)NULL;
    WideIntType.dupIntRepProc    = (Tcl_DupInternalRepProc *)NULL;
    WideIntType.updateStringProc = UpdateStringOfWideInt;
    WideIntType.setFromAnyProc   = SetWideIntFromAny;
#endif

    Tcl_RegisterObjType(&ISupportsType);
    Tcl_RegisterObjType(&NsIDType);
    Tcl_RegisterObjType(&PointerType);
    Tcl_RegisterObjType(&IdentifierType);

#ifndef USE_TCL_WIDEINT
    Tcl_RegisterObjType(&WideIntType);
#endif

}

