# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is TclXPCOM.
# 
# The Initial Developer of the Original Code is Mark Follett.
# Portions created by Mark Follett are Copyright (C) 2001-2002
# Mark Follett.  All Rights Reserved.
# 
# Contributor(s):
#     Mark Follett <mef123@myrealbox.com> (Original Author)
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.

package require tclxpcom

package provide tclxpcom_loader 0.1

namespace eval ::loader {
  variable _clsid            {{b8883660-9d40-11d5-9be9-00609717670b}}
  variable _description      "TclXPCOM Tcl Component Loader"
  variable _contractid       "@tclxpcom.org/moz.tclloader;1"
  variable _interfaces       nsIComponentLoader

  variable mimetype          "script/tcl"
  variable keyname           "software/mozilla/XPCOM/components"
  variable lastModValueName  "FileSize"
  variable fileSizeValueName "LastModTimeStamp"

  variable cmanager
  variable modules
  variable registry
  variable registry_key


  proc getModule {file} {
    variable modules
    set path [$file set path]
    if {![info exists modules($path)]} {
      set modules($path) [xpcom::module [list source $path]]
    }
    set modules($path)
  }


  proc getRegistryLocation {location} {
    variable registry

    set e_location [$registry escapeKey $location true]
    if {[xpcom::isnull $e_location]} {
      set e_location $location
    }
    return $e_location
  }


  proc fileHasChanged {location file} {
    variable registry
    variable registry_key
    variable lastModValueName
    variable fileSizeValueName

    if {![info exists registry_key]} {
      return 1
    }

    set location [getRegistryLocation $location]
    set key [$registry getSubtreeRaw $registry_key $location]

    if {[$file set lastModifiedTime] != [$registry getLongLong $key $lastModValueName]} {
      return 1
    }

    if {[$file set fileSize] != [$registry getLongLong $key $fileSizeValueName]} {
      return 1
    }

    return 0
  }


  proc setFileInfo {location file} {
    variable registry
    variable registry_key
    variable lastModValueName
    variable fileSizeValueName

    if {![info exists registry_key]} {
      return
    }

    set location [getRegistryLocation $location]
    set key [$registry addSubtreeRaw $registry_key $location]
    $registry setLongLong $key $lastModValueName  [$file set lastModifiedTime]
    $registry setLongLong $key $fileSizeValueName [$file set fileSize]
  }


  proc clearFileInfo {location} {
    variable registry
    variable registry_key

    if {![info exists registry_key]} {
      return
    }

    set location [getRegistryLocation $location]
    catch {$registry removeSubtreeRaw $registry_key $e_location}
  }


###########################################
#### Implementation of nsIComponentLoader
###########################################
  proc getFactory {clsid location type} {
    variable mimetype
    variable modules
    variable cmanager

    if {![string equal $type $mimetype]} {
      xpcom::throw $xpcom::NS_ERROR_FAILURE
    }
    puts "Getting Factory for $clsid at $location"

    set file [$cmanager nsIComponentManagerObsolete::specForRegistryLocation $location]
    set module [getModule $file]
    $module getClassObject $cmanager $clsid nsIFactory
  }


  proc init {compmanager _registry} {
    variable cmanager
    variable keyname
    variable registry
    variable registry_key

    puts "Initializing Tcl Component Loader"
    set cmanager $compmanager
    set registry [$_registry QueryInterface nsIRegistry]
    catch {set registry_key [$registry getSubTree [xpcom::constant nsIRegistry Common] $keyname]}
  }


  proc onRegister {clsid type description contractid location replace persist} {
    puts "onregister called"
  }


  proc autoRegisterComponents {when directory} {
    puts "Autoregistering Tcl Components in: [$directory set path]"
    set files [xpcom::getSimpleEnumeration [$directory set directoryEntries] nsIFile]
    foreach file $files {
      set path [$file set path]
      if {[string equal [file extension $path] ".tcl"]} {
        xpcom::invoke -noscript [xpcom::this] autoRegisterComponent $when $file
      }
    }
  }


  proc autoRegisterComponent {when file} {
    variable cmanager
    variable mimetype
    puts "Autoregistering Tcl Component: [$file set path]"

    set location [$cmanager nsIComponentManagerObsolete::registryLocationForSpec $file]
    if {![fileHasChanged $location $file]} {
      return true
    }

    set module [getModule $file]
    $module registerSelf $cmanager $file $location $mimetype
    setFileInfo $location $file
    return true
  }


  proc autoUnregisterComponent {when file} {
    variable cmanager
    puts "Autounregistering Tcl Component: [$file set path]"

    set location [$cmanager nsIComponentManagerObsolete::registryLocationForSpec $file]
    set module [getModule $file]
    $module unregisterSelf $cmanager $file $location
    clearFileInfo $location
    return true
  }


  proc registerDeferredComponents {when} {
    puts "Registering Deferred Tcl Components"
    return false
  }


  proc unloadAll {when} {
    puts "Unloading all Tcl Components"
    catch {unset modules}
  }
}


proc registerSelf   {cmanager location loaderstring type} {
  puts "Registering Tcl Component Loader"
  set categorymanager [xpcom::new -service "@mozilla.org/categorymanager;1" nsICategoryManager]
  $categorymanager addCategoryEntry "component-loader" $::loader::mimetype $::loader::_contractid true true
}

proc unregisterSelf {cmanager location loaderstring} {
  puts "Unregistering Tcl Component Loader"
  set categorymanager [xpcom::new -service "@mozilla.org/categorymanager;1" nsICategoryManager]
  $categorymanager removeCategoryEntry "component-loader" $::loader::mimetype true
}

namespace eval ::xpcom {}

proc ::xpcom::bgerror {msg} {
  global errorCode errorInfo
  puts "Error in TclXPCOM Component Loader Module: (Code: $errorCode) $msg"
  puts "$errorInfo"
}

xpcom::register                                           \
  $::loader::_clsid                                       \
  $::loader::_description                                 \
  $::loader::_contractid                                  \
  $::loader::_interfaces                                  \
  {xpcom::bindings nsIComponentLoader ::loader}

xpcom::registermodule registerSelf unregisterSelf

