/* -*- Mode: C++; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is TclXPCOM.
 * 
 * The Initial Developer of the Original Code is Mark Follett.
 * Portions created by Mark Follett are Copyright (C) 2001-2002
 * Mark Follett.  All Rights Reserved.
 * 
 * Contributor(s):
 *     Mark Follett <mef123@myrealbox.com> (Original Author)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 */

/**
 *  @file TclXPCOMLoader.cpp
 *  Contains functions that implement the TclXPCOM Tcl component
 *  loader.
 */

#ifndef MOZILLA_STRICT_API
#define MOZILLA_STRICT_API
#endif

#include "tclxpcom.h"
#include "nsIModule.h"
#include "nsIComponentManager.h"
#include "nsIFile.h"

static nsIID nsimodule_iid = NS_IMODULE_IID;

static int loaderFailed = 0;

#define LOADER_SCRIPT "                                  \
  package require tclxpcom;                              \
  xpcom::module {package require tclxpcom_loader} {      \
    puts \"unloading\"; unload                           \
  }                                                      \
"

/**
 *  Retrieves an nsIModule interface pointer from the given
 *  Tcl Object. If an error occurs and 'interp' is not NULL,
 *  a message will be left in 'interp's result.
 *  
 *  @retval TCL_OK    - if successful
 *  @retval TCL_ERROR - if an error occurred
 */
int
GetModuleFromObj(
    Tcl_Interp *interp, ///< (in)  interpreter to store any error messages
    Tcl_Obj    *objPtr, ///< (in)  Tcl object
    nsIModule **module  ///< (out) address to store interface pointer
)
{
    int rv;
    nsresult res;
    nsISupports *isupports;
    Tcl_ObjType *isupportstype;

    // no precondition for interp
    NS_PRECONDITION(objPtr, "null pointer");
    NS_PRECONDITION(module, "null pointer");

    isupportstype = Tcl_GetObjType(TCLXPCOM_ISUPPORTSTYPENAME);

    rv = Tcl_ConvertToType(interp, objPtr, isupportstype);

    if (rv == TCL_OK)
    {
        isupports = (nsISupports *)Tcl_GetHashValue((Tcl_HashEntry *)objPtr->internalRep.twoPtrValue.ptr2);

        res = isupports->QueryInterface(nsimodule_iid, (void **)module);
        if (NS_FAILED(res))
        {
            rv = TCL_ERROR;
        }
    }

    return rv;
}


/**
 *  Processes the 'unload' command, which terminates the currently
 *  running Tcl interpreter.
 *  
 *  @retval TCL_OK    - if successful
 *  @retval TCL_ERROR - if an error occurred
 */
int
UnloadObjCmd(
    ClientData  clientData,  ///< (in) contains the interpreter to release
    Tcl_Interp *interp,      ///< (in) interpreter where the command was invoked
    int         objc,        ///< (in) number of Tcl arguments
    Tcl_Obj    *CONST objv[] ///< (in) Tcl argument array
)
{
    Tcl_Interp *gInterp;

    NS_PRECONDITION(clientData, "null pointer");
    NS_PRECONDITION(interp, "null pointer");
    NS_PRECONDITION((objc >= 0), "objc must be positive");
    NS_PRECONDITION(objv, "null pointer");

    if (objc != 1)
    {
        Tcl_WrongNumArgs(interp, 1, objv, "");
        return TCL_ERROR;
    }

    gInterp = (Tcl_Interp *)clientData;

    Tcl_Preserve(gInterp);
    if (!Tcl_InterpDeleted(gInterp))
    {
        Tcl_DeleteInterp(gInterp);
    }
    Tcl_Release(gInterp);

    return TCL_OK;
}


/**
 *  Module entry point for the TclXPCOM Tcl Component Loader.
 *  
 *  @return an XPCOM result code
 */
extern "C" NS_EXPORT
nsresult
NSGetModule(
    nsIComponentManager *servMgr, ///< (in) service manager
    nsIFile             *location,///< (in) location
    nsIModule          **result   ///< (in) address to store interface pointer
)
{
    int rv;
    Tcl_Interp *interp;

    NS_PRECONDITION(servMgr, "null pointer");
    NS_PRECONDITION(location, "null pointer");
    NS_PRECONDITION(result, "null pointer");

    if (loaderFailed)
    {
        return NS_ERROR_FAILURE;
    }

    printf("Attempting to load the TclXPCOM Component Loader module...\n");

    // XXX this probably doesn't work on unix
    Tcl_FindExecutable("");

    interp = Tcl_CreateInterp();
    Tcl_Preserve(interp);

    rv = Tcl_Init(interp);
    if (rv == TCL_OK)
    {
        Tcl_CreateObjCommand(interp, "unload", UnloadObjCmd, interp, NULL);
        rv = Tcl_EvalObjEx(interp, Tcl_NewStringObj(LOADER_SCRIPT, -1), TCL_EVAL_GLOBAL | TCL_EVAL_DIRECT);
    }

    if (rv == TCL_OK)
    {
        rv = GetModuleFromObj(interp, Tcl_GetObjResult(interp), result);
    }

    if (rv != TCL_OK)
    {
        printf("Unable to load module:\n%s\n", Tcl_GetVar(interp, "errorInfo", TCL_GLOBAL_ONLY));
    }

    if (rv == TCL_OK)
    {
        // clear the module reference out of the interpreters result
        Tcl_ResetResult(interp);
        Tcl_Release(interp);
        return NS_OK;
    } else
    {
        if (!Tcl_InterpDeleted(interp))
        {
            Tcl_DeleteInterp(interp);
        }
        Tcl_Release(interp);

        loaderFailed = 1;
        return NS_ERROR_FAILURE;
    }
}

